"""
Functions to compute the loss-to-exposure ratio.
"""

from typing import Optional

import numpy as np
import pandas as pd

from . import date_to_period

REQUIRED_COLUMNS = ["actual_payout", "payout", "start", "expired_on", "expiration"]

from .. import DAY, today


def current_value(data: pd.DataFrame, post_mortem: bool = False, **kwargs) -> float:
    """
    Computes the current loss-to-exposure ratio. If post_mortem is True, the loss-to-exposure ratio is computed only on
    policies with expiration date <= today.
    """
    _data = data.copy()
    if post_mortem:
        _data = _data.loc[_data.expiration <= pd.to_datetime("today").normalize()]

    mask = ~_data.actual_payout.isna()
    loss_to_exposure = _data.loc[mask, "actual_payout"].sum() / _data.loc[mask, "payout"].sum()
    loss_to_exposure *= 100

    return loss_to_exposure


def at_t(data: pd.DataFrame, date: pd.Timestamp, post_mortem: bool = False, **kwargs) -> float:
    """
    Computes the loss-to-exposure ratio at time t. If post_mortem is True, the loss-to-exposure ratio is computed only
    on policies with expiration date <= t.
    """
    if post_mortem is True:
        mask = data.expiration < date
    else:
        mask = np.ones(len(data), dtype=bool)

    # Remove active policies
    mask = mask & ~data.actual_payout.isna()
    # Compute loss-to-exposure ratio
    if not any(mask):
        return 0
    x = data.loc[mask, "actual_payout"].sum() / data.loc[mask, "payout"].sum()
    if np.isnan(x):
        x = 0
    return 100 * x


def time_series(
    data,
    cumulative: bool = False,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: Optional[str] = "expiration",
    percent: Optional[bool] = False,
    **kwargs,
) -> pd.Series:
    """
    Computes data's loss to exposure time series. Policies are divided in different batches
    based on their expected (period_column = "expiration") or actual (period_column = "expired_on") expiration date.
    For each batch, the function computes policies' loss to exposure. If cumulative is True, the loss to exposure
    is cumulative.
    :param data: policies dataframe
    :param cumulative: if True, the loss ratio is cumulative (i.e., computed on all the policies up to the date)
    :param freq: frequency of the time series
    :param end_date: maximum date. If None, the maximum date in the period column is used
    :param start_date: minimum date. If None, the minimum date in the period column is used
    :param period_column: can be "start", "expired_on", "expiration".
    :param percent: if True, the loss to exposure is expressed in percentage points
    """
    assert "payout" in data.columns, "payout column is required"
    assert "actual_payout" in data.columns, "actual_payout column is required"
    assert period_column in data.columns, "period column is required"

    data = data[~data[period_column].isna()].copy()

    data["date"] = date_to_period(dates=data[period_column], freq=freq)
    data = data.groupby("date").agg({"payout": "sum", "actual_payout": "sum"})

    if cumulative is True:
        data["payout"] = data.payout.cumsum()
        data["actual_payout"] = data.actual_payout.cumsum()

    data["loss_to_exposure"] = data.actual_payout / data.payout

    if (percent is False) or (percent is None):
        data["loss_to_exposure"] *= 100

    if start_date is not None:
        data = data.loc[data.index >= start_date]
    if end_date is not None:
        data = data.loc[data.index <= end_date]
    return data["loss_to_exposure"]


def rolling_at_t(data: pd.DataFrame, date: pd.Timestamp, timedelta=pd.Timedelta(days=7), **kwargs) -> float:
    """
    Computes the loss-to-exposure ratio of the policies with expiration date in a given time window.
    :param data: policies dataframe
    :param date: center of the time window
    :param timedelta: half-width of the time window
    """
    mask = (data.expiration > date - timedelta) & (data.expiration <= date + timedelta)
    mask &= ~data.actual_payout.isna()
    if not any(mask):
        return 0
    x = data.loc[mask, "actual_payout"].sum() / data.loc[mask, "payout"].sum()
    if np.isnan(x):
        x = 0
    return 100 * x



"""
Utils to fetch data from Ensuro's API
"""

from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
import requests
import yaml

from . import get_from_quote

RISK_MODULES_API_ADDRESS = "https://offchain-v2.ensuro.co/api/riskmodules/"


class OffchainAPI:
    """
    Class for interacting with the offchain API.

    It provides low-level primitives (GET) for a single call, and high-level methods for combining multiple calls.
    """

    def __init__(self, url, page_size=1000, private_key=None):
        self.url = url.strip("/")
        self.page_size = page_size
        self.private_key = private_key
        self.session = requests.Session()
        if private_key:
            self.session.headers.update({"Authorization": f"Token {private_key}"})

    def __repr__(self):
        return f"<OffchainAPI url={self.url}>"

    def url_to_endpoint(self, url):
        """
        Returns the endpoint for the given URL. Useful for links returned by the API.
        """
        if url.startswith(self.url):
            url = url[len(self.url) :]
        return url.strip("/")

    def endpoint_url(self, endpoint):
        """
        Returns the full URL for the given endpoint.
        """
        if endpoint.startswith(self.url):
            endpoint = endpoint[len(self.url) :]
        return f"{self.url}/{endpoint.strip('/')}/"

    def get(self, endpoint, query_params: dict = None, validate_response=True):
        """
        Low level GET call to the given endpoint.
        """
        response = self.session.get(self.endpoint_url(endpoint), params=query_params)
        if validate_response:
            response.raise_for_status()
        return response

    def multi_page_get(self, endpoint, query_params: dict = None):
        """
        Low level GET call to the given endpoint, paginating through all the results.
        """
        url = self.endpoint_url(endpoint) + f"?limit={self.page_size}"
        if query_params:
            url += "&" + "&".join([f"{k}={v}" for k, v in query_params.items()])

        results = []
        while url is not None:
            response = self.session.get(url)
            response.raise_for_status()
            results += response.json()["results"]
            url = response.json()["next"]

        return results

    def get_policies(self):
        """
        Returns a list with all the policies.
        """
        policies = []
        url = self.endpoint_url("portfolio") + f"?limit={self.page_size}"
        if self.private_key is not None:
            url += "&private_data=True"

        while url is not None:
            response = self.session.get(url)
            response.raise_for_status()
            policies += response.json()["results"]
            url = response.json()["next"]

        return policies

    @staticmethod
    def get_expiration_date(row, version="v2"):
        """
        Computes the actual expiration date for a policy
        :param row: row of the policy dataset as it comes from API
        :param version: dataset from API V1 or API V2
        :return: actual expiration date
        """
        if version == "v1":
            colname = "tx"
        elif version == "v2":
            colname = "event"
        else:
            raise ValueError

        transactions_list = row[f"{colname}s"]
        for transaction in transactions_list:
            if transaction[f"{colname}_type"] == "resolution":
                return transaction["timestamp"]
        return np.nan

    def build_policies_table(self, quote_columns: Optional[str | Path] = None):
        """
        Downloads data from Ensuro's API and transforms it in the "standard" Ensuro policies table.
        """
        # Get Policies Data
        policies = self.get_policies()
        # create dataset
        df = pd.DataFrame.from_records(policies).fillna(np.nan)
        to_numeric = [
            "payout",
            "premium",
            "jr_scr",
            "sr_scr",
            "loss_prob",
            "pure_premium",
            "ensuro_commission",
            "partner_commission",
            "jr_coc",
            "sr_coc",
            "actual_payout",
        ]
        df[to_numeric] = df[to_numeric].astype(float)

        # Transform data
        now = pd.to_datetime("today")

        expiration_dates = [self.get_expiration_date(row, version="v2") for row in policies]
        df["expired_on"] = expiration_dates
        df.loc[df.actual_payout == 0, "expired_on"] = df.loc[df.actual_payout == 0, "expiration"].values

        # format dates
        df["start"] = pd.to_datetime(df.start).dt.tz_localize(None)
        df["expiration"] = pd.to_datetime(df.expiration).dt.tz_localize(None)
        df["expired_on"] = pd.to_datetime(df.expired_on).dt.tz_localize(None)
        df["active"] = df.actual_payout.apply(np.isnan)

        df["progress"] = (now - df.start).apply(lambda x: np.timedelta64(x) / np.timedelta64(1, "s"))
        df["progress"] = df["progress"] / (df.expiration - df.start).apply(
            lambda x: np.timedelta64(x) / np.timedelta64(1, "s")
        )
        df.loc[~df.active, "progress"] = 1.0

        df["duration_expected"] = (df.expiration - df.start).apply(lambda x: x / np.timedelta64(1, "D"))
        df["duration_actual"] = (df.expired_on - df.start).apply(lambda x: x / np.timedelta64(1, "D"))

        # create risk module
        df["risk_module"] = df["rm"].apply(lambda x: x.split("/")[-2])

        df["partner"] = [get_from_quote(x, "partnerName") for x in df.quote]
        df["ttype"] = [get_from_quote(x, "ticketType") for x in df.quote]

        if quote_columns is not None:
            # Create columns out of quote
            with open(quote_columns, "r") as f:
                quote_columns_dict = yaml.safe_load(f)

            for key, columns in quote_columns_dict.items():
                for column in columns:
                    df[column] = [get_from_quote(x, column) for x in df.quote]

        df["start_date"] = pd.to_datetime(df.start).dt.date

        risk_modules = requests.get(RISK_MODULES_API_ADDRESS)
        risk_modules = risk_modules.json()
        risk_modules = dict(zip([x["address"] for x in risk_modules], [x["name"] for x in risk_modules]))

        df["rm_name"] = df.risk_module.map(risk_modules)

        df["start_date"] = pd.to_datetime(df["start_date"])
        df["date_price_applied"] = (
            df.groupby(["rm_name", "loss_prob"]).start.transform("min").apply(lambda x: str(x.date()))
        )

        return df



"""
Utils to download ensuro data
"""

import json
from typing import List, Optional

import pandas as pd
from google.cloud import bigquery
from google.oauth2 import service_account

from ensuro_analytics.download import ENSURO_API_URL
from ensuro_analytics.download.api import OffchainAPI
from ensuro_analytics.download.utils import etoken_blockshot_processing as etk_processing

DEFAULT_POLICIES_TABLE_COLUMNS = [
    "id",
    "ensuro_id",
    "payout",
    "loss_prob",
    "jr_scr",
    "sr_scr",
    "pure_premium",
    "ensuro_commission",
    "partner_commission",
    "jr_coc",
    "sr_coc",
    "start",
    "expiration",
    "actual_payout",
    "expired_on",
    "premium",
    "active",
    "progress",
    "duration_expected",
    "duration_actual",
    "risk_module",
    "rm_name",
    "quote",
]

DATETIME_COLUMNS = ["date", "start", "expiration", "expired_on"]


class BigQueryInterface:
    """
    A class used to interact with the Google BigQuery API.

    Attributes
    ----------
    project_id : str
        the Google Cloud project ID
    dataset_name : str
        the name of the dataset in BigQuery
    account_key_path : Optional[str]
        the path to the service account key file
    policies_table_columns : Optional[list[str]]
        the list of columns to fetch from the policies table
    credentials : google.auth.credentials.Credentials
        the credentials to use for the BigQuery client
    Client : google.cloud.bigquery.client.Client
        the BigQuery client

    Methods
    -------
    _date_cols_to_datetime(df: pd.DataFrame, columns: List[str] = DATETIME_COLUMNS) -> pd.DataFrame:
        Converts the specified columns of the dataframe to datetime.
    _bytes_to_dict(col: pd.Series) -> pd.Series:
        Converts a series of bytes to a series of dictionaries.
    sql(sql_query: str) -> pd.DataFrame:
        Executes a SQL query and returns the result as a dataframe.
    policies_table(limit: Optional[int] = None) -> pd.DataFrame:
        Fetches data from the policies table and returns it as a dataframe.
    time_series_table() -> pd.DataFrame:
        Fetches data from the time series table and returns it as a dataframe.
    fetch_data(table: str) -> pd.DataFrame:
        Fetches data from the specified table and returns it as a dataframe.
    """

    def __init__(
        self,
        project_id: str,
        dataset_name: str,
        account_key_path: Optional[str] = None,
        policies_table_columns: Optional[list[str]] = None,
    ):
        """
        Constructs a new BigQueryInterface object.

        Parameters
        ----------
        project_id : str
            the Google Cloud project ID
        dataset_name : str
            the name of the dataset in BigQuery
        account_key_path : Optional[str]
            the path to the service account key file
        policies_table_columns : Optional[list[str]]
            the list of columns to fetch from the policies table
        """
        self.project_id = project_id
        self.dataset_name = dataset_name
        self.account_key_path = account_key_path
        self.credentials = None

        if self.account_key_path is not None:
            self.credentials = service_account.Credentials.from_service_account_file(self.account_key_path)

        self.Client = bigquery.Client(credentials=self.credentials, project=self.project_id)

        if policies_table_columns is None:
            self.policies_table_columns = DEFAULT_POLICIES_TABLE_COLUMNS
        else:
            self.policies_table_columns = policies_table_columns

    @staticmethod
    def _date_cols_to_datetime(df: pd.DataFrame, columns: List[str] = DATETIME_COLUMNS):
        """
        Converts the specified columns of the dataframe to datetime.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe whose columns are to be converted.
        columns : List[str], optional
            The list of column names to convert, by default DATETIME_COLUMNS

        Returns
        -------
        pd.DataFrame
            The dataframe with the specified columns converted to datetime.
        """
        for col in columns:
            if col in df.columns:
                df[col] = pd.to_datetime(df[col])
        return df

    @staticmethod
    def _bytes_to_dict(col: pd.Series) -> pd.Series:
        """
        Converts a series of bytes to a series of dictionaries.

        Parameters
        ----------
        col : pd.Series
            The pandas Series containing bytes to be converted.

        Returns
        -------
        pd.Series
            The pandas Series where each byte has been converted to a dictionary.
        """
        quote_types = col.apply(type).unique()
        if bytes in quote_types:
            col = col.apply(lambda x: x.decode("utf-8") if isinstance(x, bytes) else x)
            col = col.apply(lambda x: json.loads(x) if isinstance(x, str) else x)
        return col

    def sql(self, sql_query: str) -> pd.DataFrame:
        """
        Executes a SQL query and returns the result as a dataframe.

        Parameters
        ----------
        sql_query : str
            The SQL query to be executed.

        Returns
        -------
        pd.DataFrame
            The result of the SQL query as a dataframe.
        """
        df = self.Client.query(sql_query).to_dataframe()
        return df

    def policies_table(self, limit: Optional[int] = None) -> pd.DataFrame:
        """
        Fetches data from the policies table and returns it as a dataframe.

        Parameters
        ----------
        limit : Optional[int]
            The maximum number of rows to return.

        Returns
        -------
        pd.DataFrame
            The data from the policies table as a dataframe.
        """
        columns_str = ", ".join(self.policies_table_columns)
        sql_query = f"SELECT * FROM `{self.project_id}.{self.dataset_name}.policies_table`"
        if limit is not None:
            sql_query += f" ORDER BY id DESC LIMIT {limit}"
        sql_query = f"SELECT {columns_str} FROM ({sql_query}) ORDER BY id ASC"

        data = self._date_cols_to_datetime(self.sql(sql_query))
        data["quote"] = self._bytes_to_dict(data.quote)

        return data

    def time_series_table(self) -> pd.DataFrame:
        """
        Fetches data from the time series table and returns it as a dataframe.

        Returns
        -------
        pd.DataFrame
            The data from the time series table as a dataframe.
        """
        sql_query = f"SELECT * FROM `{self.project_id}.{self.dataset_name}.time_series`"
        return self._date_cols_to_datetime(self.sql(sql_query))

    def etoken_blockshot(self) -> pd.DataFrame:
        """
        Fetches data from the etoken_blockshot table and returns it as a dataframe.

        Returns
        -------
        pd.DataFrame
            The data from the etoken_blockshot table as a dataframe.
        """
        sql_query = f"SELECT * FROM `{self.project_id}.{self.dataset_name}.etoken_block_shot`"
        return self._date_cols_to_datetime(self.sql(sql_query))

    def token_metrics(self, include_insurance_returns: bool = False) -> pd.DataFrame:
        """
        Fetches data and builds the etoken_metrics dataframe.

        Parameters
        ----------
        include_insurance_returns : bool
            If True, computes the returns generated by Ensuro's insurance activity; If False, only the compound returns
            of Ensuro's insurance activity and investments are returned.
        Returns
        -------
        pd.DataFrame
            The data from the token_metrics table as a dataframe.
        """

        etoken_blockshot = self.etoken_blockshot()

        offchain_api = OffchainAPI(ENSURO_API_URL)

        lps = pd.json_normalize(offchain_api.multi_page_get("lpevents"), sep="_")
        etokens_api_query = offchain_api.get("etokens").json()

        token_metrics = etk_processing.blocks_shots_to_token_metrics(
            etoken_blockshot, lps, etokens_api_query
        )

        if include_insurance_returns is True:
            riskmodules, sr_etks_to_rm, jr_etks_to_rm = etk_processing.get_etokens_to_risk_modules_map(
                offchain_api
            )
            policies = self.policies_table()
            insurance_returns = etk_processing.build_insurance_returns(
                token_metrics, policies, sr_etks_to_rm, jr_etks_to_rm
            )
            token_metrics["dividend_insurance"] = insurance_returns.dividend_insurance

        return token_metrics

    def fetch_data(self, table: str) -> pd.DataFrame:
        """
        Fetches data from the specified table and returns it as a dataframe.

        Parameters
        ----------
        table : str
            The name of the table to fetch data from. Must be either 'portfolio' or 'time-series'.

        Returns
        -------
        pd.DataFrame
            The data from the specified table as a dataframe.

        Raises
        ------
        AssertionError
            If the table name is not 'portfolio' or 'time-series'.
        """
        assert table in [
            "portfolio",
            "time-series",
        ], "table must be either 'portfolio' or 'time-series'"

        if table == "portfolio":
            data = self.policies_table()
        elif table == "time-series":
            data = self.time_series_table()

        return data



"""
Functions to compute premium balance.
"""

from typing import Optional

import numpy as np
import pandas as pd

from .. import DAY, today
from . import date_to_period

REQUIRED_COLUMNS = [
    "premium",
    "pure_premium",
    "actual_payout",
    "expiration",
    "expired_on",
    "start",
]


def current_value(
    data: pd.DataFrame,
    post_mortem: bool = False,
    use_pure_premium: bool = True,
    **kwargs,
) -> float:
    """
    Computes the current premium balance. If post_mortem is True, the premium balance is computed only on
    policies with expiration date <= today. If use_pure_premium is True, the pure premium is used instead of the
    premium.
    :param data: policies dataframe
    :param post_mortem: If True, the premium balance is computed only on policies with expiration date <= today
    :param use_pure_premium: If True, the pure premium is used instead of the premium. Default is True
    """
    _data = data.copy()
    if post_mortem:
        _data = _data.loc[data.expiration <= pd.to_datetime("today").normalize()]
    premium_column = "pure_premium" if use_pure_premium else "premium"
    premium_balance = _data.loc[:, premium_column].sum() - np.nansum(_data.loc[:, "actual_payout"])
    return premium_balance


def at_t(
    data: pd.DataFrame,
    date: pd.Timestamp,
    post_mortem: bool = False,
    use_pure_premium: bool = True,
    **kwargs,
) -> float:
    """
    Computes the premium balance at time t. If post_mortem is True, the premium balance is computed only on
    policies with expiration date <= t. If use_pure_premium is True, the pure premium is used instead of the
    premium.
    :param data: policies dataframe
    :param date: time t
    :param post_mortem: If True, the premium balance is computed only on policies with expiration date <= t
    :param use_pure_premium: If True, the pure premium is used instead of the premium. Default is True
    """
    if use_pure_premium is True:
        premium_column = "pure_premium"
    else:
        premium_column = "premium"

    mask_premium = data.start <= date
    mask_payout = data.expired_on <= date

    if post_mortem is True:
        mask_premium &= data.expiration <= date
        mask_payout &= data.expiration <= date

    premium_balance = (
        data.loc[mask_premium, premium_column].sum() - data.loc[mask_payout].actual_payout.sum()
    )
    return premium_balance


def time_series(
    data,
    cumulative: bool = False,
    freq: str = "1W",
    use_pure_premium: bool = True,
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: str = "expiration",
    **kwargs,
) -> pd.Series:
    """
    Computes the time series of premium balance for the data. Policies are divided in different batches
    based on their expected (period_column = "expiration") or actual (period_column = "expired_on") expiration date.
    For each batch, the function computes policies' premium balance.
    If cumulative is True, the premium balance is cumulative.
    :param data: policies dataframe
    :param cumulative: if True, the premium balance is cumulative (i.e., computed on all the policies up to the date)
    :param freq: frequency of the time series
    :param use_pure_premium: If True, the pure premium is used instead of the premium. Default is True
    :param end_date: maximum date. If None, the maximum date in the period column is used
    :param start_date: minimum date. If None, the minimum date in the period column is used
    :param period_column: column to use for expiration date. It can be "expiration", "expired_on".
    """
    if use_pure_premium is True:
        premium_column = "pure_premium"
    else:
        premium_column = "premium"

    assert premium_column in data.columns, f"{premium_column} column is required"
    assert "actual_payout" in data.columns, "actual_payout column is required"
    assert period_column in data.columns, "expiration column is required"

    _data = data.copy()

    # Transform dates in regular intervals
    _data["date"] = date_to_period(dates=_data[period_column], freq=freq)
    _data = _data.groupby("date").agg({premium_column: "sum", "actual_payout": lambda x: np.nansum(x)})

    if cumulative:
        _data[premium_column] = _data[premium_column].cumsum()
        _data["actual_payout"] = _data.actual_payout.cumsum()

    _data["premium_balance"] = _data[premium_column] - _data["actual_payout"]

    if start_date is not None:
        _data = _data.loc[_data.index >= start_date]
    if end_date is not None:
        _data = _data.loc[_data.index <= end_date]

    return _data["premium_balance"]



from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio


def _colorway() -> list:
    """
    Returns the list of colors from the plotly colorway
    :return: list of colors
    """
    return pio.templates[pio.templates.default].layout.colorway


def plot(
    fig: go.Figure,
    metric: pd.Series,
    cumulative: bool,
    label: str,
    volume: Optional[pd.Series] = None,
    benchmark: Optional[float] = None,
    percent_axis: Optional[bool] = True,
) -> go.Figure:
    """
    Generates the time series plot
    :param fig: plotly figure
    :param metric: series containing the data to plot
    :param cumulative: Whether to plot cumulative or rolling series
    :param label: label for the y axis
    :param volume: series containing the volume to plot
    :param benchmark: benchmark to plot
    :param percent_axis: whether to plot the y axis as percentage
    :return: a plotly figure
    """

    if cumulative:
        title = f"Cumulative {label}".title()
    else:
        title = f"Rolling {label}".title()

    fig.add_trace(
        go.Scatter(
            x=metric.index,
            y=metric.values,
            name=title,
            mode="lines+markers",
            marker={
                "color": "white",
                "size": 10,
                "line": {"color": _colorway()[0], "width": 3},
            },
            line={"color": _colorway()[0]},
        ),
        secondary_y=False,
    )

    if volume is not None:
        fig.add_trace(
            go.Scatter(
                x=volume.index,
                y=volume.values,
                name="Volume",
                mode="lines+markers",
                marker={
                    "color": "white",
                    "size": 10,
                    "line": {"color": _colorway()[1], "width": 2},
                    "symbol": "diamond",
                },
                line={"color": _colorway()[1]},
                opacity=0.7,
            ),
            secondary_y=True,
        )

    if benchmark is not None:
        fig.add_trace(
            go.Scatter(
                x=metric.index,
                y=np.ones(len(metric.index)) * benchmark,
                hoverinfo="skip",
                mode="lines",
                line={"color": "red", "dash": "dash"},
                name="Benchmark",
            ),
            secondary_y=False,
        )

    if percent_axis:
        fig.update_yaxes(tickformat="p", secondary_y=False)

    if volume is not None:
        fig.update_yaxes(title="Volume", secondary_y=True, showgrid=False, side="right")

    fig.update_yaxes(
        secondary_y=False,
        title=title,
        showgrid=False,
        side="left",
        title_font=dict(color=_colorway()[0]),
    )
    fig.update_xaxes(title="Date")
    fig.update_layout(showlegend=False)

    if volume is not None:
        fig.update_yaxes(title_font=dict(color=_colorway()[1]), secondary_y=True)

    fig.update_layout(hovermode="x", hoverlabel=dict(font=dict(color="white"), bgcolor=_colorway()[0]))

    return fig


def plot_by_primary_channel(
    fig: go.Figure,
    metric_dic: dict,
    cumulative: bool,
    label: str,
    benchmark: Optional[float] = None,
    percent_axis: Optional[bool] = True,
) -> go.Figure:
    """
    Generates the time series plot by primary channel
    :param fig: plotly figure
    :param metric_dic: dictionary containing the data to plot
    :param cumulative: Whether to plot cumulative or rolling series
    :param label: label for the y axis
    :param benchmark: benchmark to plot
    :param percent_axis: whether to plot the y axis as percentage
    :return: a plotly figure
    """

    if cumulative:
        title = f"Cumulative {label}".title()
    else:
        title = f"Rolling {label}".title()

    for i, (channel, metric) in enumerate(metric_dic.items()):
        fig.add_trace(
            go.Scatter(
                x=metric.index,
                y=metric.values,
                name=channel,
                mode="lines+markers",
                marker={
                    "color": "white",
                    "size": 10,
                    "line": {"color": _colorway()[i], "width": 3},
                },
                line={"color": _colorway()[i]},
            ),
            secondary_y=False,
        )

    all_indices = pd.concat([metric for (_, metric) in metric_dic.items()]).index
    idx = pd.Index([all_indices.min(), all_indices.max()])

    if benchmark is not None:
        fig.add_trace(
            go.Scatter(
                x=idx,
                y=np.ones(len(idx)) * benchmark,
                hoverinfo="skip",
                mode="lines",
                line={"color": "red", "dash": "dash"},
                name="Benchmark",
            ),
            secondary_y=False,
        )

    if percent_axis:
        fig.update_yaxes(tickformat="p", secondary_y=False)

    fig.update_yaxes(title=title, secondary_y=False, showgrid=False, side="left")
    fig.update_xaxes(title="Date")
    fig.update_layout(showlegend=True)

    return fig



"""
This module contains functions to compute post mortem premium.
"""

import warnings

import pandas as pd

REQUIRED_COLUMNS = ["expiration", "pure_premium"]


def current_value(data: pd.DataFrame, **kwargs) -> float:
    """
    Computes the current overall post mortem premium.
    """
    return at_t(
        data,
        pd.to_datetime("today"),
    )


def at_t(data: pd.DataFrame, date: pd.Timedelta, **kwargs) -> float:
    """
    Computes the post mortem premium at time t.
    """
    mask = data.expiration < date
    return data.loc[mask].pure_premium.sum()


# TODO: scrivere questa funzione
def time_series(data, freq="1W", end_date="today", **kwargs):
    """
    Computes the time series of post mortem premium for the data. Policies are divided in different batches
    based on their expected expiration date. For each batch, the function computes the post mortem premium.
    """
    warnings.warn(
        "post_mortem_premium.time_series not implemented",
        UserWarning,
    )
    pass



from importlib import import_module
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objects as go

from ..analytics import DAY, _timestamp, today
from ..analytics.portfolio.total_policies import time_series as total_policies_time_series
from .matplotlib_layout import time_series as matplotlib_layout
from .plotly_layout import time_series as plotly_layout

AVAILABLE_METRICS = [
    "active_policies",
    "active_premium",
    "collected_premium",
    "default_percentage",
    "junior_scr",
    "loss_ratio",
    "loss_to_exposure",
    "n_payouts",
    "payouts",
    "premium_balance",
    "scr",
    "senior_scr",
    "total_policies",
    "var",
]

METRICS_PATH = "..analytics.portfolio.{}"


def plot(
    data: pd.DataFrame,
    metric: str,
    fig: Union[go.Figure, plt.Figure],
    cumulative: Optional[bool] = True,
    freq: Optional[str] = "1W",
    period_column: Optional[str] = "expiration",
    post_mortem: Optional[bool] = True,
    with_volume: Optional[bool] = True,
    benchmark: Optional[float] = None,
    percent: Optional[bool] = False,
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    **kwargs,
) -> Union[go.Figure, plt.Figure]:
    """
    Generates the metric plot
    :param data: Dataframe containing the data
    :param metric: Metric to plot
    :param fig: Figure object
    :param cumulative: Whether to plot cumulative or rolling loss ratio
    :param freq: Temporal frequency
    :param period_column: Name of the column to use as time
    :param post_mortem: Whether to plot only policies with expiration <= max expired_on
    :param with_volume: Whether to plot volume
    :param benchmark: Benchmark value to plot
    :param percent: Whether to plot as percentage
    :param end_date: End date of the plot
    :param start_date: Start date of the plot
    :return: Plotly or matplotlib figure depending on layout parameter
    """

    assert (
        metric in AVAILABLE_METRICS
    ), f"Metric {metric} not available. Available metrics: {AVAILABLE_METRICS}"

    data = data.copy()

    module = import_module(METRICS_PATH.format(metric), package="ensuro_analytics.visual")

    ts = module.time_series(
        data,
        cumulative=cumulative,
        freq=freq,
        end_date=end_date,
        start_date=start_date,
        period_column=period_column,
        port_mortem=post_mortem,
        percent=True,
        **kwargs,
    )

    if with_volume:
        volume = total_policies_time_series(
            data,
            cumulative=cumulative,
            freq=freq,
            end_date=end_date,
            start_date=start_date,
            port_mortem=post_mortem,
            period_column=period_column,
        )
    else:
        volume = None

    if isinstance(fig, go.Figure):
        return plotly_layout.plot(
            fig,
            ts,
            cumulative,
            label=metric.replace("_", " "),
            volume=volume,
            benchmark=benchmark,
            percent_axis=percent,
        )
    elif isinstance(fig, plt.Figure):
        return matplotlib_layout.plot(
            fig,
            ts,
            cumulative,
            label=metric.replace("_", " "),
            volume=volume,
            benchmark=benchmark,
            percent_axis=percent,
        )


def plot_by_primary_channel(
    data: pd.DataFrame,
    metric: str,
    fig: Union[go.Figure, plt.Figure],
    primary_channel: Optional[str] = "partner",
    cumulative: Optional[bool] = True,
    freq: Optional[str] = "1W",
    period_column: Optional[str] = "expiration",
    post_mortem: Optional[bool] = True,
    benchmark: Optional[float] = None,
    n_channel: Optional[int] = 5,
    percent: Optional[bool] = False,
    end_date: Optional[pd.Timestamp | str] = today(),
    start_date: Optional[pd.Timestamp | str] = today() - 90 * DAY,
) -> Union[go.Figure, plt.Figure]:
    """
    Generates the metric plot by primary channel
    :param data: Dataframe containing the data
    :param metric: Metric to plot
    :param fig: Figure object
    :param primary_channel: Name of the primary channel column to plot
    :param cumulative: Whether to plot cumulative or rolling loss ratio
    :param freq: Temporal frequency
    :param period_column: Name of the column to use as time
    :param post_mortem: Whether to plot only policies with expiration <= max expired_on
    :param benchmark: Benchmark value to plot
    :param n_channel: Number of channels to plot
    :param percent: Whether to plot as percentage
    :param end_date: End date of the plot
    :param start_date: Start date of the plot
    :return: Plotly or matplotlib figure depending on layout parameter
    """

    assert (
        metric in AVAILABLE_METRICS
    ), f"Metric {metric} not available. Available metrics: {AVAILABLE_METRICS}"

    # Turn start_date and end_date to datetime
    if isinstance(start_date, str):
        start_date = _timestamp(start_date)
    if isinstance(end_date, str):
        end_date = _timestamp(end_date)

    _data = data.copy()

    module = import_module(METRICS_PATH.format(metric), package="ensuro_analytics.visual")

    mask = np.ones(_data.shape[0], dtype=bool)
    if start_date is not None:
        mask &= _data[period_column] >= start_date
    if end_date is not None:
        mask &= _data[period_column] <= end_date

    channels = _data[mask][primary_channel].value_counts().head(n_channel).index
    ts = {}

    for ch in channels:
        ts[ch] = module.time_series(
            _data[_data[primary_channel] == ch],
            cumulative=cumulative,
            freq=freq,
            end_date=end_date,
            start_date=start_date,
            period_column=period_column,
            percent=percent,
            post_mortem=post_mortem,
        )

    if isinstance(fig, go.Figure):
        return plotly_layout.plot_by_primary_channel(
            fig,
            ts,
            cumulative,
            label=metric.replace("_", " "),
            benchmark=benchmark,
            percent_axis=percent,
        )
    elif isinstance(fig, plt.Figure):
        return matplotlib_layout.plot_by_primary_channel(
            fig,
            ts,
            cumulative,
            label=metric.replace("_", " "),
            benchmark=benchmark,
            percent_axis=percent,
        )



"""
Utils to create time-series data from a table of policies.
"""

import pandas as pd

from .portfolio.active_policies import at_t as active_policies_at_t
from .portfolio.active_premium import at_t as active_premium_at_t
from .portfolio.collected_premium import at_t as won_premium_at_t
from .portfolio.collected_premium import rolling_at_t as won_premium_rolling_at_t
from .portfolio.default_percentage import at_t as default_percentage_at_t
from .portfolio.default_percentage import rolling_at_t as default_percentage_rolling_at_t
from .portfolio.loss_ratio import at_t as loss_ratio_at_t
from .portfolio.loss_ratio import rolling_at_t as loss_ratio_rolling_at_t
from .portfolio.n_payouts import at_t as n_payouts_at_t
from .portfolio.payouts import at_t as payout_at_t
from .portfolio.post_mortem_premium import at_t as post_mortem_premium_at_t
from .portfolio.scr import at_t as scr_at_t
from .portfolio.total_policies import at_t as total_policies_at_t
from .portfolio.total_policies import rolling_at_t as total_policies_rolling_at_t


def compute_aggregate_metrics(data, date):
    """
    Compute several different analytics metrics at time t.
    """
    active_premium = active_premium_at_t(data, date)
    inactive_premium = won_premium_at_t(data, date)
    active_policies = active_policies_at_t(data, date)
    total_policies = total_policies_at_t(data, date)
    total_payouts = payout_at_t(data, date)
    num_payouts = n_payouts_at_t(data, date)
    total_scr_used = scr_at_t(data, date)
    loss_ratio = loss_ratio_at_t(data, date)
    rolling_loss_ratio = loss_ratio_rolling_at_t(data, date)
    default_percentage = default_percentage_at_t(data, date)
    rolling_default_percentage = default_percentage_rolling_at_t(data, date)
    rolling_n_policies = total_policies_rolling_at_t(data, date)
    rolling_premium = won_premium_rolling_at_t(data, date)
    post_mortem_premium = post_mortem_premium_at_t(data, date)

    results = (
        active_premium,
        inactive_premium,
        active_policies,
        total_policies,
        total_payouts,
        num_payouts,
        total_scr_used,
        loss_ratio,
        rolling_loss_ratio,
        default_percentage,
        rolling_default_percentage,
        rolling_n_policies,
        rolling_premium,
        post_mortem_premium,
    )

    return results


def build_time_series_dataframe(data, freq="1W"):
    """
    Function to aggregate the policy dataset in a time series of metrics. The metrics
    produced are indexed by date/product
    Parameters:
    data (pd.DataFrame): policies dataframe
    risk_module_map (dict): dictionary mapping smart contract addresses to risk modules names
    Returns
    df_ts (pd.DataFrame): dataframe with time series metrics.
    """

    start_date = data.start.min().normalize()
    end_date = pd.to_datetime("today").normalize()

    dates = pd.date_range(start=start_date, end=end_date, normalize=True, freq=freq)
    time_series = pd.DataFrame(dates, columns=["date"])

    get_data = lambda: data.copy()

    # Fill values
    new_columns = [
        "rm_name",
        "active_premium",
        "inactive_premium",
        "active_policies",
        "total_policies",
        "total_payouts",
        "num_payouts",
        "total_scr_used",
        "loss_ratio",
        "rolling_loss_ratio",
        "default_perc",
        "rolling_default_perc",
        "rolling_n_policies",
        "rolling_premium",
        "post_mortem_premium",
    ]

    time_series[new_columns] = pd.DataFrame.from_records(
        time_series["date"].apply((lambda x: compute_aggregate_metrics(get_data(), x)))
    ).values

    return time_series



"""
Functions to compute exposure.
"""

from typing import Optional

import pandas as pd

from .. import DAY, active_at_t, today
from . import _max_date_range, date_to_period

REQUIRED_COLUMNS = ["payout", "start", "expiration", "expired_on"]


def current_value(data: pd.DataFrame, **kwargs) -> float:
    """
    Computes the current exposure.
    """
    mask = active_at_t(data, today())
    return data.loc[mask].payout.sum()


def at_t(data: pd.DataFrame, date: pd.Timestamp, **kwargs) -> float:
    """
    Computes the exposure at time t.
    """
    mask = active_at_t(data, date)
    return data.loc[mask].payout.sum()


def time_series(
    data: pd.DataFrame,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: str = "start",
    **kwargs,
) -> pd.Series:
    """
    Computes the time series of exposure for the data.
    :param data: policies dataframe
    :param freq: time frequency of the time series
    :param end_date: maximum date. If None, the maximum date in the period column is used
    :param start_date: minimum date. If None, the minimum date in the period column is used
    :param period_column: can be "start", "expired_on",  "expiration", or "max".
    :return:
    """
    assert "expired_on" in data.columns, "expired_on column is required"
    assert "start" in data.columns, "start column is required"

    # Get unique dates at a regular interval
    if period_column == "max":
        dates = _max_date_range(data=data, freq=freq)
    else:
        dates = sorted(date_to_period(dates=data[period_column], freq=freq).unique())

    exposure = []
    get_data = lambda: data[["start", "expired_on", "payout"]].copy()
    # Compute active policies at each date
    for date in dates:
        exposure.append(at_t(get_data(), date))

    exposure = pd.Series(exposure, index=dates, name="exposure")

    # Filter by start and end date
    if start_date is not None:
        exposure = exposure.loc[exposure.index >= start_date]
    if end_date is not None:
        exposure = exposure.loc[exposure.index <= end_date]

    return exposure



from dataclasses import dataclass
from functools import lru_cache

import numpy as np
import pandas as pd
import requests
import yfinance as yf
from environs import Env

env = Env()
env.read_env()

API_V2_URL = env.str("API_V2_URL", "https://offchain-v2.ensuro.co/api/")
LPS_URL = env.str("ETOKEN_URL", API_V2_URL + "lpevents/")


@dataclass
class Today:
    """
    Class to handle the current date
    """

    @property
    def date(self):
        return pd.Timestamp("today").normalize()

    @property
    def str(self):
        return self.date.strftime("%Y-%m-%d")

    def back(self, days: int):
        return (self.date - pd.Timedelta(days=days)).date().strftime("%Y-%m-%d")


today = Today()


class TimeUnits:
    """
    Class to handle time units.
    """

    def __init__(self, unit: str):
        assert unit in ["1D", "1W", "1M", "1Y"], "Invalid unit"
        self.unit = unit

    @property
    def n_units_in_one_year(self):
        """
        Returns the number of units in one year.
        """

        match self.unit:
            case "1D":
                return 365
            case "1W":
                return 52.14
            case "1M":
                return 12.17
            case "1Y":
                return 1
            case _:
                return 0

    @property
    def n_days_in_unit(self):
        """
        Returns the number of days in one unit.
        """

        match self.unit:
            case "1D":
                return 1
            case "1W":
                return 7
            case "1M":
                return 30
            case "1Y":
                return 365
            case _:
                return 0

    @property
    def daily_freq(self):
        """
        Returns the number of days in one unit.
        """

        match self.unit:
            case "1D":
                return "1D"
            case "1W":
                return "7D"
            case "1M":
                return "30D"
            case "1Y":
                return "365D"
            case _:
                return "1D"


@lru_cache
def risk_free_rate(
    start_date: str | pd.Timestamp = today.back(365),
    end_date: str | pd.Timestamp = today.str,
):
    """
    Function to get the risk-free rate from treasury bills.
    :param start_date: Start date for the risk-free rate.
    :param end_date: End date for the risk-free rate.
    :return risk_free_rate: The risk-free rate.
    """

    if isinstance(start_date, pd.Timestamp):
        start_date = start_date.strftime("%Y-%m-%d")

    if isinstance(end_date, pd.Timestamp):
        end_date = end_date.strftime("%Y-%m-%d")

    # Download the 3-month treasury bills
    risk_free_rate = yf.download("^IRX", start=start_date, end=end_date)
    risk_free_rate = risk_free_rate.reset_index()["Open"].mean() / 100

    return risk_free_rate


@lru_cache
def get_market_data(
    start_date: str = today.back(3650),
    end_date: str = today.str,
    ticker: str = "^GSPC",
) -> pd.DataFrame:
    """
    Function to get the market data.
    :param start_date: Start date for the market data.
    :param end_date: End date for the market data.
    :return market_data: The market data.
    """

    # Download the S&P500 data
    df = yf.download(ticker, start=start_date, end=end_date).reset_index()
    df["Date"] = df["Date"].dt.tz_localize("UTC").dt.normalize()
    df.rename(columns={"Close": "close", "Date": "date"}, inplace=True)
    df.set_index("date", inplace=True)

    return df[["close"]]


def get_market_returns(
    dates=None,
    ticker: str = "^GSPC",
    start_date: str = None,
    end_date: str = None,
    freq: str = "1W",
) -> pd.DataFrame:
    """
    Function to get the market returns.
    Either dates or (start_date, end_date, freq) must be provided.
    If dates are provided, the market data is matched with the list of dates.
    If start_date, end_date and freq are provided, the market data is matched with the dates generated from the start_date, end_date and freq.
    :param dates: Dates for the market data.
    :param ticker: Ticker for the market data.
    :param start_date: Start date for the market data.
    :param end_date: End date for the market data.
    :param freq: Frequency for the market data.
    :return market_data: The market data.
    """

    market_data = get_market_data(ticker=ticker)

    # Match the market data with the eToken data, in terms of dates
    # In the market data some dates may be missing due to weekends or holidays
    market_data = _match_market_etoken_dates(dates, start_date, end_date, freq, market_data)

    return market_data[["date", "returns_sp"]]


def _match_market_etoken_dates(dates, start_date, end_date, freq, market_data):
    if dates is not None:
        market_dates = pd.Series(dates)
        market_mask = ~market_dates.isin(market_data.index)
        max_shift = 0
        n_shift = market_mask.sum()
        while market_mask.any():
            market_dates.loc[market_mask] = market_dates.loc[market_mask] - pd.Timedelta(days=1)
            market_mask = ~market_dates.isin(market_data.index)
            max_shift += 1
        if max_shift > 0:
            print(f"Shifted {n_shift} market data entries up to {max_shift} days to match the eToken data.")
        market_data = market_data.loc[market_dates]
        market_data["returns_sp"] = market_data["close"].pct_change()
        market_data.index = dates
        market_data.reset_index(inplace=True)

    elif start_date is not None and end_date is not None:
        dates = pd.date_range(start=start_date, end=end_date, freq=freq)
        market_data = market_data.loc[dates]

        market_data["returns_sp"] = market_data["close"].pct_change()
        market_data.reset_index(inplace=True)

    else:
        raise ValueError("Either dates or start_date and end_date must be provided.")
    return market_data


@lru_cache
def market_returns(
    start_date: str = today.back(365),
    end_date: str = today.str,
    time_resolution: str | TimeUnits = "1W",
    ticker: str = "^GSPC",
) -> pd.DataFrame:
    """
    Function to prepare the S&P500 data to be used in the metrics.
    :param start_date: Start date for the S&P500 data.
    :param end_date: End date for the S&P500 data.
    :param time_resolution: Time resolution for the S&P500 data.
    """

    if isinstance(time_resolution, str):
        time_resolution = TimeUnits(time_resolution)

    start_date = (pd.Timestamp(start_date) - pd.Timedelta(days=365)).strftime("%Y-%m-%d")

    # Download the S&P500 data
    sp_df = yf.download(ticker, start=start_date, end=end_date).reset_index()
    sp_df["Date"] = sp_df["Date"].dt.tz_localize("UTC").dt.normalize()
    sp_df.rename(columns={"Close": "close", "Date": "date"}, inplace=True)
    sp_df.set_index("date", inplace=True)

    sp_df = sp_df.resample(time_resolution.daily_freq).agg({"close": "last"})
    sp_df["returns_sp"] = sp_df["close"].pct_change()

    # Drop the last date (may have incomplete data)
    sp_df = sp_df.iloc[:-1]
    sp_df.reset_index(inplace=True)

    return sp_df[["date", "returns_sp"]]


def get_lps():
    """Function to get the LPs data from the API."""

    session = requests.Session()
    lps_list = []
    url = LPS_URL
    while True:
        response = session.get(url)
        lps_list += response.json()["results"]
        if response.json()["next"] is None:
            break
        url = response.json()["next"]

    return pd.json_normalize(lps_list, sep="_")


def blocks_shots_to_token_metrics(
    etoken_blocks: pd.DataFrame, lps_df: pd.DataFrame, etokens_api_query: list[dict]
) -> pd.DataFrame:
    """
    Function to transform the eToken blocks shots and the LPs data into a DataFrame with daily resolution and all the necessary columns.
    :param etoken_blocks: DataFrame with the eToken blocks shots.
    :param lps_df: DataFrame with the LPs data.
    """

    # Transform the eToken blocks shots columns with correct types
    etoken_blocks["total_supply"] = etoken_blocks.total_supply.astype(float)
    etoken_blocks["scr"] = etoken_blocks.scr.astype(float)
    etoken_blocks["date"] = pd.to_datetime(etoken_blocks["date"], utc=True).dt.normalize()

    # Transform the LPs data columns with correct types
    lps_df.event_tx_timestamp = pd.to_datetime(lps_df.event_tx_timestamp)
    lps_df["date"] = lps_df.event_tx_timestamp.dt.normalize()

    # Compute the deposits and withdrawals
    lps_df["deposit"] = (lps_df.event_type == "deposit").astype(int) * lps_df.amount.astype(float)
    lps_df["withdraw"] = (lps_df.event_type == "withdraw").astype(int) * lps_df.amount.astype(float)

    # Get the eToken address and the eToken URL
    etokens_url_address_map = {r["url"]: r["address"] for r in etokens_api_query}

    # Transform the eToken URL to the eToken address
    lps_df["e_token"] = lps_df.e_token.map(etokens_url_address_map)

    # Compute the daily token balance
    daily_token_balance = (
        etoken_blocks.sort_values(["e_token", "date"])
        .groupby(["e_token", "date"])
        .agg(
            {
                "total_supply": "last",
                "scr": "last",
            }
        )
    )

    # Set a multiindex from the eToken address and the date
    e_tokens = daily_token_balance.reset_index().e_token.unique()
    dates = daily_token_balance.reset_index().date.unique()

    # Don't consider the very last date (may have incomplete data)
    dates = pd.date_range(start=dates.min(), end=dates.max(), freq="D")[:-1]

    # Set the multiindex
    reindex = pd.MultiIndex.from_product([e_tokens, dates], names=["e_token", "date"])
    daily_token_balance = daily_token_balance.reindex(reindex)

    # Fill the missing values of total supply and SCR with the last available value
    daily_token_balance = daily_token_balance.groupby(level="e_token").transform(lambda x: x.bfill().ffill())

    # Compute daily deposits and withdrawals and fill the missing values with 0
    daily_token_balance["deposit"] = lps_df.groupby(["e_token", "date"]).deposit.sum()
    daily_token_balance["withdraw"] = lps_df.groupby(["e_token", "date"]).withdraw.sum()
    daily_token_balance[["withdraw", "deposit"]] = daily_token_balance[["withdraw", "deposit"]].fillna(0)

    # Compute dividends; The dividends are computed as the difference between the total supply of the eToken
    # at time t+1 and t, minus the deposits and plus the withdrawals.
    daily_token_balance["dividend"] = (
        daily_token_balance.groupby(level="e_token").total_supply.diff()
        - daily_token_balance.deposit
        + daily_token_balance.withdraw
    )

    return daily_token_balance


def returns_dataframe(etokens_data: pd.DataFrame, time_resolution: str = "1D") -> pd.DataFrame:
    """
    Computes the dataframe of returns, utilization rate, flow, outlay, and proceeds for each eToken (both nominal and perfect).
    :param etokens_data: Dataframe with eTokens' total supply, deposits, withdrawals, scr, and dividends.
    :param time_resolution: Time resolution for the returns.
    """

    if isinstance(time_resolution, str):
        time_resolution = TimeUnits(time_resolution)

    _aggregation_funcs = {
        "total_supply": "last",
        "scr": "last",
        "dividend": "sum",
        "deposit": "sum",
    }

    # Resample with desired time resolution
    returns = (
        etokens_data.reset_index(level="e_token")
        .groupby("e_token")
        .resample(time_resolution.daily_freq)
        .agg(_aggregation_funcs)
    )

    # Compute the returns
    returns["returns"] = (
        returns["dividend"] / returns.groupby(level="e_token").total_supply.shift()
    ).replace(np.inf, np.nan)
    returns["perfect_returns"] = (
        returns["dividend"] / returns.groupby(level="e_token").scr.shift()
    ).replace(np.inf, np.nan)

    # Compute the utilization rate
    returns["UR"] = returns[f"scr"] / returns["total_supply"]

    # Compute the flow
    returns["flow"] = returns["total_supply"].groupby(level="e_token").diff(1) - returns["dividend"]
    returns["perfect_flow"] = returns["scr"].groupby(level="e_token").diff(1)

    # Fill the first occurrence with the total supply (or SCR)
    first_occurrences = returns.groupby(level="e_token").cumcount() == 0
    returns.loc[first_occurrences, "flow"] = returns.loc[first_occurrences, "total_supply"]
    returns.loc[first_occurrences, "perfect_flow"] = returns.loc[first_occurrences, "scr"]

    # Compute the cumulative flow
    returns["cumflow"] = returns["flow"].cumsum()
    returns["perfect_cumflow"] = returns["perfect_flow"].cumsum()

    # Compute the outlay and proceeds
    returns["outlay"] = returns["flow"].apply(lambda x: 0 if x < 0 else x)
    returns["proceeds"] = -returns["flow"].apply(lambda x: 0 if x > 0 else x)
    returns["perfect_outlay"] = returns["perfect_flow"].apply(lambda x: 0 if x < 0 else x)
    returns["perfect_proceeds"] = -returns["perfect_flow"].apply(lambda x: 0 if x > 0 else x)

    # Drop the last aggregate date (may have incomplete data)
    dates = returns.index.levels[1]
    returns = returns.loc[(slice(None), dates[:-1]), :]

    # Clean the index

    return returns


def filter_first_nonzero(group: pd.DataFrame, column="deposit") -> pd.DataFrame:
    """
    Function to filter the first occurrence of a column with a nonzero value.
    :param group: DataFrame group.
    :param column: Column to filter.
    """
    nonzero_index = group.reset_index()[column].ne(0).idxmax() + 1

    return group.iloc[nonzero_index:]


def filter_returns_dataframe(returns: pd.DataFrame, filter_method="first_deposit"):
    if filter_method == "first_deposit":
        # Start the time series only after the first deposit
        returns = (
            returns.groupby(level="e_token").apply(filter_first_nonzero).reset_index(level=0, drop=True)
        )
    else:
        raise ValueError(f"Filter method {filter_method} not implemented")

    return returns



"""
Functions to compute the collected premium.
"""

from typing import Optional

import pandas as pd

from .. import DAY, active_at_t, today
from . import _max_date_range, date_to_period

REQUIRED_COLUMNS = ["premium", "start", "expired_on"]


def current_value(data: pd.DataFrame, **kwargs) -> float:
    """
    Computes the current value of the collected premium
    """
    mask = active_at_t(data, pd.to_datetime("today"))
    return data.loc[~mask].premium.sum()


def at_t(data: pd.DataFrame, date: pd.Timestamp, **kwargs) -> float:
    """
    Computes the value of the collected premium at a given date
    """
    mask = (~data.expired_on.isna()) & (data.expired_on < date)
    return data.loc[mask].premium.sum()


def time_series(
    data: pd.DataFrame,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: str = "start",
    **kwargs,
) -> pd.Series:
    """
    Computes the time series of the collected ("won") premiums at a given date.
    :param data: policies dataframe
    :param freq: time frequency of the time series
    :param end_date: maximum date. If None, the maximum date in the period column is used
    :param start_date: minimum date. If None, the minimum date in the period column is used
    :param period_column: can be "start", "expired_on", "expiration", or "max".
    """
    assert "premium" in data.columns, "premium column is required"
    assert period_column in data.columns, f"{period_column} column is required"
    assert "expired_on" in data.columns, "expired_on column is required"

    if period_column == "max":
        dates = _max_date_range(data=data, freq=freq)
    else:
        dates = sorted(date_to_period(dates=data[period_column], freq=freq).unique())

    inactive_premiums = []

    get_data = lambda: data[["expired_on", "premium"]].copy()

    for date in dates:
        inactive_premiums.append(at_t(get_data(), date))

    inactive_premiums = pd.Series(inactive_premiums, index=dates, name="collected_premiums")

    if start_date is not None:
        inactive_premiums = inactive_premiums.loc[inactive_premiums.index >= start_date]
    if end_date is not None:
        inactive_premiums = inactive_premiums.loc[inactive_premiums.index <= end_date]

    return inactive_premiums


def rolling_at_t(data: pd.DataFrame, date: pd.Timestamp, timedelta=pd.Timedelta(days=7)):
    """
    Computes the value of the collected premium for policies expired in a given time window.
    :param data: policies dataframe
    :param date: center of the time window
    :param timedelta: half-width of the time window; default is 7 days.
    """
    data = data.loc[~data.expired_on.isna()]
    mask = (data.expired_on > date - timedelta) & (data.expired_on <= date + timedelta)
    return data.loc[mask].pure_premium.sum()



"""
Functions to compute the loss ratio.
"""

from typing import Optional

import numpy as np
import pandas as pd

from .. import DAY, today
from . import date_to_period

REQUIRED_COLUMNS = ["pure_premium", "actual_payout", "expiration", "expired_on"]


def current_value(data: pd.DataFrame, post_mortem: bool = False, **kwargs) -> float:
    """
    Computes the current loss ratio. If post_mortem is True, the loss ratio is computed only on policies with expiration
    date <= today.
    """
    _data = data.copy()
    if post_mortem:
        _data = _data.loc[data.expiration <= pd.to_datetime("today").normalize()]
    loss_ratio = _data.actual_payout.sum() / _data.pure_premium.sum() * 100
    return loss_ratio


def at_t(data: pd.DataFrame, date: pd.Timestamp, post_mortem: bool = False, **kwargs) -> float:
    """
    Computes the loss ratio at time t. If post_mortem is True, the loss ratio is computed only on policies with
    expiration date <= t.
    """

    if post_mortem is True:
        mask = data.expiration < date
    else:
        mask = np.ones(len(data), dtype=bool)
    # Remove active policies
    mask &= ~data.actual_payout.isna()

    if not any(mask.values):
        return 0
    # Compute loss ratio
    x = data.loc[mask].actual_payout.sum() / data.loc[mask].pure_premium.sum()
    if np.isnan(x):
        x = 0
    return 100 * x


def time_series(
    data,
    cumulative: bool = False,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: Optional[str] = "expiration",
    percent: Optional[bool] = False,
    **kwargs,
) -> pd.Series:
    """
    Computes data's loss ratio time series. Policies are divided in different batches
    based on their expected (period_column = "expiration") or actual (period_column = "expired_on") expiration date.
    For each batch, the function computes policies' default percentage. If cumulative is True, the loss ratio
    is cumulative.
    :param data: policies dataframe
    :param cumulative: if True, the loss ratio is cumulative (i.e., computed on all the policies up to the date)
    :param freq: frequency of the time series
    :param end_date: maximum date. If None, the maximum date in the period column is used
    :param start_date: minimum date. If None, the minimum date in the period column is used,
    :param period_column: can be "expiration", "expired_on".
    :param percent: if True, the loss ratio is expressed as a percentage
    """
    assert "pure_premium" in data.columns, "pure_premium column is required"
    assert "actual_payout" in data.columns, "actual_payout column is required"
    assert period_column in data.columns, "period column is required"

    data = data[~data[period_column].isna()].copy()

    data["date"] = date_to_period(dates=data[period_column], freq=freq)
    data = data.groupby("date").agg({"pure_premium": "sum", "actual_payout": "sum"})

    if cumulative:
        data["pure_premium"] = data.pure_premium.cumsum()
        data["actual_payout"] = data.actual_payout.cumsum()

    data["loss_ratio"] = data.actual_payout / data.pure_premium
    if not percent:
        data["loss_ratio"] *= 100

    if start_date is not None:
        data = data.loc[data.index >= start_date]
    if end_date is not None:
        data = data.loc[data.index <= end_date]
    return data["loss_ratio"]


def rolling_at_t(data, date, timedelta=pd.Timedelta(days=7)) -> float:
    """
    Computes the loss ratio of the policies with expiration date in a given time window.
    :param data: policies dataframe
    :param date: center of the time window
    :param timedelta: half-width of the time window
    """
    mask = (data.expiration > date - timedelta) & (data.expiration <= date + timedelta)
    mask = mask & ~data.actual_payout.isna()
    x = data.loc[mask].actual_payout.sum() / data.loc[mask].pure_premium.sum()
    if np.isnan(x):
        x = 0
    return 100 * x



"""
Common function for the analytics
"""

from typing import Any

import numpy as np
import pandas as pd

DAY = pd.Timedelta("1D")


def today() -> pd.Timestamp:
    return pd.to_datetime("today").normalize()


def _timestamp(date: Any) -> pd.Timestamp:
    return pd.Timestamp(date).normalize()


def active_at_t(data: pd.DataFrame, date: pd.Timestamp) -> pd.Series:
    mask = data.expired_on.isna() | (data.expired_on > _timestamp(date))
    mask &= data.start <= _timestamp(date)
    return mask


def _between(
    dates: pd.DataFrame,
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    include_left: bool = True,
    include_right: bool = True,
) -> pd.Series:
    """
    Finds dates between a start and an end date
    :param dates:
    :param start_date:
    :param end_date:
    :param include_left:
    :param include_right:
    :return:
    """
    mask = np.ones(len(dates)).astype(bool)

    if include_left is True:
        mask &= dates >= start_date
    else:
        mask &= dates > start_date

    if include_right is True:
        mask &= dates <= end_date
    else:
        mask &= dates < end_date

    return mask


def expired_between(
    data: pd.DataFrame,
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    include_left: bool = True,
    include_right: bool = True,
) -> pd.Series:
    """
    Finds policies expired between two dates
    :param data: policies table
    :param start_date:
    :param end_date:
    :param include_left: include start date in the interval
    :param include_right: include end date in the interval
    """

    return _between(
        data.expired_on,
        start_date=start_date,
        end_date=end_date,
        include_left=include_left,
        include_right=include_right,
    )


def started_between(
    data: pd.DataFrame,
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    include_left: bool = True,
    include_right: bool = True,
) -> pd.Series:
    """
    Finds policies started between two dates
    :param data: policies table
    :param start_date:
    :param end_date:
    :param include_left: include start date in the interval
    :param include_right: include end date in the interval
    """

    return _between(
        data.start,
        start_date=start_date,
        end_date=end_date,
        include_left=include_left,
        include_right=include_right,
    )


def _count_payouts(x):
    return ((~np.isnan(x)) & (x > 1e-9)).sum()


def find_first_date(
    data: pd.DataFrame, splitters=list[str], date_column: str = "start", **kwargs
) -> pd.Series:
    """
    Finds the date of the first policy sold for each of the groups identified by the splitters
    """
    sorted_data = data.copy().sort_values(by=date_column)
    first_dates = sorted_data.groupby(splitters)[date_column].transform("min").dt.date

    return first_dates.reindex(data.index).copy()



"""
This module contains the functions to download data from Ensuro's API and BigQuery.
"""

ENSURO_API_URL = "https://offchain-v2.ensuro.co/api/"


def get_from_quote(x, field):
    """
    This function attempts to retrieve a specific field from policies' metadata in the 'quote' column.

    Parameters:
    x (dict): The quote from which to retrieve the field. It is expected to be a dictionary with a "data" key.
    field (str): The name of the field to retrieve from the quote.

    Returns:
    The value of the specified field if it exists, None otherwise.
    """
    try:
        return x["data"][field]
    except:
        return None



import warnings
from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.ticker import FuncFormatter


def _colorway() -> list:
    """
    Returns the list of colors from the plotly colorway
    :return: list of colors
    """
    return matplotlib.rcParams["axes.prop_cycle"].by_key()["color"]


def percentage_formatter(x, pos):
    return f"{x*100:.1f}%"


def plot(
    fig: plt.Figure,
    metric: pd.Series,
    cumulative: bool,
    label: str,
    volume: Optional[pd.Series] = None,
    benchmark: Optional[float] = None,
    percent_axis: Optional[bool] = True,
) -> plt.Figure:
    """
    Generates the time series plot
    :param fig: matplotlib figure
    :param metric: series containing the data to plot
    :param cumulative: Whether to plot cumulative or rolling series
    :param label: label for the y axis
    :param volume: series containing the volume to plot
    :param benchmark: benchmark to plot
    :param percent_axis: whether to plot the y axis as percentage
    :return: a matplotlib figure
    """

    if cumulative:
        title = f"Cumulative {label}".title()
    else:
        title = f"Rolling {label}".title()

    ax1 = fig.gca()

    ax1.plot(
        metric.index,
        metric.values,
        color=_colorway()[0],
        markerfacecolor="w",
        marker="o",
        markersize=12,
        lw=2,
        markeredgewidth=4,
        label=title,
    )
    ax1.set_ylabel(title, color=_colorway()[0])

    if volume is not None:
        ax2 = ax1.twinx()
        ax2.plot(
            volume.index,
            volume.values,
            markerfacecolor="w",
            markersize=12,
            lw=2,
            markeredgewidth=4,
            color=_colorway()[1],
            marker="D",
            label="Volume",
            alpha=0.7,
        )
        ax2.set_ylabel("Volume", color=_colorway()[1])
        ax2.spines["right"].set_visible(True)

    if benchmark is not None:
        ax1.axhline(
            benchmark,
            color="red",
            linestyle="--",
            linewidth=2,
            label="Benchmark",
        )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if percent_axis:
            ax1.yaxis.set_major_formatter(FuncFormatter(percentage_formatter))

    # set the dates in the x-axis in the format MM-DD
    ax1.xaxis.set_major_formatter(matplotlib.dates.DateFormatter("%d-%m"))
    ax1.set_xlabel("Date")
    fig.tight_layout()

    plt.close()

    return fig


def plot_by_primary_channel(
    fig: plt.Figure,
    metric_dic: dict,
    cumulative: bool,
    label: str,
    benchmark: Optional[float] = None,
    percent_axis: Optional[bool] = True,
) -> plt.Figure:
    """
    Generates the time series plot by primary channel
    :param fig: matplotlib figure
    :param metric_dic: dictionary containing the data to plot
    :param cumulative: Whether to plot cumulative or rolling series
    :param label: label for the y axis
    :param benchmark: benchmark to plot
    :param percent_axis: whether to plot the y axis as percentage
    :return: a matplotlib figure
    """

    if cumulative:
        title = f"Cumulative {label}".title()
    else:
        title = f"Rolling {label}".title()

    ax1 = fig.gca()

    for i, (channel, metric) in enumerate(metric_dic.items()):
        ax1.plot(
            metric.index,
            metric.values,
            color=_colorway()[i],
            markerfacecolor="w",
            marker="o",
            markersize=12,
            lw=2,
            markeredgewidth=4,
            label=channel,
        )

    if benchmark is not None:
        ax1.axhline(
            benchmark,
            color="red",
            linestyle="--",
            linewidth=2,
            label="Benchmark",
        )

    ax1.set_ylabel(title)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if percent_axis:
            ax1.yaxis.set_major_formatter(FuncFormatter(percentage_formatter))

    # set the dates in the x-axis in the format MM-DD
    ax1.xaxis.set_major_formatter(matplotlib.dates.DateFormatter("%d-%m"))
    ax1.set_xlabel("Date")

    ax1.legend(frameon=False, loc=(1.05, 0.5))

    fig.tight_layout()

    plt.close()

    return fig



"""
Functions to compute the overall number of policies sold.
"""

from typing import Optional

import pandas as pd

from .. import DAY, _timestamp, started_between, today
from . import date_to_period

REQUIRED_COLUMNS = ["expiration", "start", "pure_premium", "premium"]


def current_value(data, post_mortem: bool = False, **kwargs) -> int:
    """
    Computes the total number of policies sold. If post_mortem is True, the number of policies sold is computed only on
    policies with expiration date <= today.
    """
    _data = data.copy()
    if post_mortem:
        _data = _data.loc[_data.expiration <= pd.to_datetime("today").normalize()]

    return _data.shape[0]


def at_t(data: pd.DataFrame, date: pd.Timestamp, post_mortem: bool = False, **kwargs) -> int:
    """
    Computes the number of policies sold at time t. If post_mortem is True, the number of policies sold is computed only on
    policies with expiration date <= t.
    """
    mask = data.start <= _timestamp(date)
    if post_mortem:
        mask &= data.expiration <= _timestamp(date)
    return mask.sum()


def time_series(
    data: pd.DataFrame,
    freq: str = "1W",
    cumulative: bool = False,
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: str = "start",
    **kwargs,
) -> pd.Series:
    """
    Computes the time series of number of policies sold for the data. Policies are divided in different batches
    based on their start date. For each batch, the function computes the number of policies sold.
    If cumulative is True, the number of policies sold is cumulative.
    :param data: policies dataframe
    :param freq: frequency of the time series
    :param cumulative: If True, the number of policies sold is cumulative
    :param end_date: maximum date of the time series. If None, the maximum date in period column is used
    :param start_date: minimum date of the time series. If None, the minimum date in period column is used
    :param period_column: column used to group policies. Can be "start", "expired_on", "expiration".
    """
    _data = data.copy()
    _data["date"] = date_to_period(dates=_data[period_column], freq=freq)
    _data = (
        _data.groupby("date")
        .agg({"pure_premium": "count"})
        .rename(columns={"pure_premium": "total_policies"})
    )

    if cumulative is True:
        _data["total_policies"] = _data["total_policies"].cumsum()

    if start_date is not None:
        _data = _data.loc[_data.index >= start_date]
    if end_date is not None:
        _data = _data.loc[_data.index <= end_date]

    return _data["total_policies"]


def rolling_at_t(data: pd.DataFrame, date: pd.Timestamp, timedelta=pd.Timedelta(days=7), **kwargs) -> int:
    """
    Computes the number of policies sold in a given time window.
    :param data: policies dataframe
    :param date: time window center
    :param timedelta: half-width of the time window
    """
    start_date = date - timedelta
    end_date = date + timedelta
    mask = started_between(data, start_date=start_date, end_date=end_date)
    return mask.sum()



"""
Functions to compute the number of payouts.
"""

from typing import Optional

import numpy as np
import pandas as pd

from .. import DAY, _count_payouts, today
from . import date_to_period

REQUIRED_COLUMNS = ["actual_payout", "start", "expiration", "expired_on"]


def current_value(data: pd.DataFrame, post_mortem: bool = False, **kwargs) -> float:
    """
    Computes the current overall number of payouts. If post_mortem is True, the number of payouts is computed only on
    policies with expiration date <= today.
    """
    _data = data.copy()
    if post_mortem:
        _data = _data.loc[_data.expiration <= pd.to_datetime("today").normalize()]

    n_payouts = (_data.actual_payout > 1e-9).sum()

    return n_payouts


def at_t(data: pd.DataFrame, date: pd.Timestamp, post_mortem: bool = False, **kwargs) -> float:
    """
    Computes the number of payouts at time t. If post_mortem is True, the number of payouts is computed only on
    policies with expiration date <= t.
    """
    if post_mortem is True:
        mask = data.expiration < date
    else:
        mask = np.ones(len(data), dtype=bool)

    mask &= data.expired_on <= date
    return (data.loc[mask].actual_payout > 1e-9).sum()


def time_series(
    data: pd.DataFrame,
    cumulative: bool = False,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: Optional[str] = "expiration",
    **kwargs,
) -> pd.Series:
    """
    Computes the time series of number of payouts for the data. Policies are divided in different batches
    based on their expected expiration date. For each batch, the function computes the number of payouts.
    If cumulative is True, the number of payouts is cumulative.
    :param data: policies dataframe
    :param cumulative: If True, the number of payouts is cumulative
    :param freq: frequency of the time series
    :param end_date: maximum date of the time series. If None, the maximum expiration date is used
    :param start_date: minimum date of the time series. If None, the minimum expiration date is used
    :param period_column: can be "start", "expired_on", "expiration".
    """
    assert "actual_payout" in data.columns, "actual_payout column is required"
    assert "start" in data.columns, "start column is required"
    assert "expired_on" in data.columns, "expired_on column is required"

    _data = data.copy()

    # Find first day of expiration week
    _data["date"] = date_to_period(dates=_data[period_column], freq=freq)

    _data = (
        _data.groupby("date")
        .agg({"actual_payout": _count_payouts})
        .rename(columns={"actual_payout": "n_payout"})
    )

    if cumulative:
        _data["n_payout"] = _data.n_payout.cumsum()

    if start_date is not None:
        _data = _data.loc[_data.index >= start_date]
    if end_date is not None:
        _data = _data.loc[_data.index <= end_date]

    return _data["n_payout"]



"""
Setup plotting libraries
"""

from pathlib import Path
from typing import Optional

import matplotlib as mpl
import plotly.graph_objects as go
import plotly.io as pio
import yaml

# Get directory of this python file

DIR = Path(__file__).parent.absolute()
VISUAL_CONFIGS_PATH = DIR / "visual_configs.yaml"

_ensuro_colors = None


def get_ensuro_colors() -> dict:
    global _ensuro_colors
    if _ensuro_colors is not None:
        return _ensuro_colors
    else:
        with open(VISUAL_CONFIGS_PATH) as f:
            _ensuro_colors = yaml.safe_load(f)["ensuro_colors"]
        return _ensuro_colors


class VizSetup:
    def __init__(self, configs: Optional[str] = None):
        self.configs = configs

    def _get_params(self) -> dict:
        """
        Loads default parameters for boh matplotlib and plotly from a yaml file
        :param configs:
        :return:
        """
        if self.configs is not None:
            with open(self.configs) as f:
                _params = yaml.safe_load(f)
            return _params
        else:
            with open(VISUAL_CONFIGS_PATH) as f:
                _params = yaml.safe_load(f)
            return _params

    def _get_mpl_params(self) -> dict:
        """
        Loads parameters for matplotlib
        """
        if self.configs is not None:
            with open(self.configs) as f:
                _params = yaml.safe_load(f)["matplotlib"]
            return _params
        else:
            with open(VISUAL_CONFIGS_PATH) as f:
                _params = yaml.safe_load(f)["matplotlib"]
            return _params

    def _get_plotly_params(self) -> dict:
        """
        Loads parameters for plotly
        """
        if self.configs is not None:
            with open(self.configs) as f:
                _params = yaml.safe_load(f)["plotly"]
            return _params
        else:
            with open(VISUAL_CONFIGS_PATH) as f:
                _params = yaml.safe_load(f)["plotly"]
            return _params

    def setup_matplotlib(self) -> None:
        # custom_rcparams = self._get_mpl_params()
        custom_rcparams = self._get_mpl_params()

        # Handle the color cycle separately
        if "axes.prop_cycle_colors" in custom_rcparams:
            colors = custom_rcparams.pop("axes.prop_cycle_colors")
            mpl.rcParams["axes.prop_cycle"] = mpl.cycler(color=colors)

        mpl.rcParams.update(custom_rcparams)
        return None

    def setup_plotly(self) -> None:
        # plotly_params = self._mpl_params_to_plotly_template()
        plotly_plain = self._get_plotly_params()
        plotly_params = dict(layout=go.Layout(plotly_plain["layout"]), data=plotly_plain["data"])
        pio.templates["ensuro"] = plotly_params
        pio.templates.default = "ensuro"
        return None

    @staticmethod
    def to_pink_blue_green_rotation() -> None:
        ensuro_colors = get_ensuro_colors()
        colors = [
            ensuro_colors["fluo"][0],
            ensuro_colors["blues"][2],
            ensuro_colors["fluo"][1],
            ensuro_colors["blues"][1],
        ]
        # For matplotlib
        mpl.rcParams["axes.prop_cycle"] = mpl.cycler(color=colors)
        # For plotly
        pio.templates["ensuro"]["layout"]["colorway"] = colors
        return None

    def to_configs_rotation(self) -> None:
        mpl_params = self._get_mpl_params()
        colors = mpl_params.get("axes.prop_cycle_colors", None)
        if colors is None:
            Warning("No color cycle found in configs")
        else:
            # For matplotlib
            mpl.rcParams["axes.prop_cycle"] = mpl.cycler(color=colors)
            # For plotly
            pio.templates["ensuro"]["layout"]["colorway"] = colors
        return None


def setup_matplotlib(configs: Optional[str] = None) -> VizSetup:
    """
    Setup matplotlib configuration
    """
    vizsetup = VizSetup(configs=configs)
    vizsetup.setup_matplotlib()
    return vizsetup


def setup_plotly(configs: Optional[str] = None) -> VizSetup:
    """
    Setup plotly configuration
    """
    vizsetup = VizSetup(configs=configs)
    vizsetup.setup_plotly()
    return vizsetup



"""
Portfolio review
"""

import warnings
from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd

from ensuro_analytics.analytics import DAY, today
from ensuro_analytics.analytics.dataframe import create_ensuro_accessors


def _get_var(x: pd.DataFrame, level: float | int | list[float] | list[int], **kwargs) -> float | list[float]:
    if isinstance(level, float | int):
        level = [level]
    try:
        return x.var.current_value(level=level, **kwargs)
    except:
        return [np.nan] * len(level)


def compute_adjusted_target_loss_ratio(
    data: pd.DataFrame,
    target_loss_ratio: float,
    tau_1: float = 30 * 3,
    tau_2: float = 30 * 3,
    lower_bound: float = 0.0,
    upper_bound: float = 1.0,
    pct_trend: float = 1.0,
    consider_active_policies: bool = True,
) -> tuple[float, float]:
    """
    Compute the adjusted target loss ratio.

    Parameters
    ----------
    data : pd.DataFrame
        The data containing the portfolio.
    target_loss_ratio : float
        The target loss ratio.
    tau_1 : float, optional
        The time period for computing the historical surplus and loss-to-exposure, by default 90 days.
    tau_2 : float, optional
        The time period for computing the expected exposure, by default 90 days. This is the period over which
        the deficit is expected to be recovered.
    lower_bound : float, optional
        The lower bound for the adjusted target loss to exposure ratio, by default 0.5.
    upper_bound : float, optional
        The upper bound for the adjusted target loss to exposure ratio, by default 0.95.
    pct_trend : float, optional
        The percentage trend for the expected exposure, by default 1.0.
    consider_active_policies : bool, optional
        Whether to consider active policies in the computation, by default True.

    Returns
    -------
    trget_loss_to_exposure, target_loss_ratio
        The adjusted target loss to exposure ratio and the adjusted target loss ratio.
    """
    # Create the ensuro accessors
    create_ensuro_accessors()

    # Compute the surplus over the last tau_1 days
    mask = (data.expiration < today()) & (data.expiration >= today() - tau_1 * DAY)
    historical_loss_to_exposure = data.loc[mask].loss_to_exposure.current_value(post_mortem=True) / 100
    # Compute the surplus
    historical_surplus = data.loc[mask, "pure_premium"].sum() - data.loc[mask, "actual_payout"].sum()

    if consider_active_policies:
        # Estimate the surplus for the active policies
        historical_surplus += (
            data.loc[data.expired_on.isna(), "pure_premium"].sum()
            - data.loc[data.expired_on.isna(), "payout"].sum() * historical_loss_to_exposure
        )

    target_loss_to_exposure = historical_loss_to_exposure / target_loss_ratio

    if historical_surplus >= 0:
        # If the performance is positive, simply return the target loss ratio
        return target_loss_to_exposure * 100, target_loss_ratio * 100

    else:
        historical_surplus *= -1

    # Compute the expected exposure that will be received in the next tau_2 days;
    # at the moment, this is set = the pure premium of the last tau_2 days
    mask = data.start >= today() - tau_2 * DAY
    expected_exposure = data.loc[mask, "payout"].sum() * pct_trend

    # We want the pure premium collected with the next tau_2 days to be enough to cover:
    # 1. the expected loss +
    # 2. the historical surplus lost in the last tau_1 days

    target_loss_to_exposure = (
        historical_loss_to_exposure / target_loss_ratio
    )  # This is the target loss to exposure
    # ratio without the deficit
    extra_loss_to_exposure = (
        historical_surplus / expected_exposure
    )  # This is the extra loss to exposure ratio

    # The target loss to exposure ratio is adjusted by the extra loss to exposure ratio
    adjusted_target_loss_to_exposure = target_loss_to_exposure + extra_loss_to_exposure

    # Bound between lower_bound and upper_bound
    adjusted_target_loss_to_exposure = np.clip(adjusted_target_loss_to_exposure, lower_bound, upper_bound)

    # Compute the adjusted target loss ratio
    adjusted_target_loss_ratio = historical_loss_to_exposure / adjusted_target_loss_to_exposure

    return adjusted_target_loss_to_exposure * 100, adjusted_target_loss_ratio * 100


@dataclass
class PortfolioReview:
    """
    A class used to review a portfolio of insurance policies.

    ...

    Attributes
    ----------
    data : pd.DataFrame
        a pandas DataFrame containing the policies data
    split_on : list[str]
        a list of columns to split the portfolio on

    Methods
    -------
    from_data(data: pd.DataFrame, split_on: str | list[str] = ["rm_name"]) -> "PortfolioReview":
        Creates a PortfolioReview object from a dataframe.
    _validate_data(data: pd.DataFrame, cols: Optional[list[str]] = None):
        Validates the data
    review(show_first_date: bool = True, show_predicted_loss_to_exposure: bool = True, show_current_portfolio_pct: bool = True, **kwargs) -> "_CompiledReview":
        Computes the portfolio review
    """

    data: pd.DataFrame
    split_on: list[str]

    @classmethod
    def from_data(
        cls,
        data: pd.DataFrame,
        split_on: str | list[str] = ["rm_name"],
        validate_columns: Optional[list[str]] = None,
    ) -> "PortfolioReview":
        """
        Creates a PortfolioReview object from a dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            policies dataframe
        split_on : str | list[str]
            list of columns to split the portfolio on
        validate_columns : Optional[list[str]]
            list of columns that the data should have; standard portfolio columns are validated by default.

        Returns
        -------
        PortfolioReview
            a PortfolioReview object
        """

        cls._validate_data(data, cols=validate_columns)
        if isinstance(split_on, str):
            split_on = [split_on]

        create_ensuro_accessors()

        return cls(
            data=data,
            split_on=split_on,
        )

    @staticmethod
    def _validate_data(data: pd.DataFrame, cols: Optional[list[str]] = None):
        """
        Validates the data

        Parameters
        ----------
        data : pd.DataFrame
            policies dataframe
        cols : Optional[list[str]]
            list of columns to validate
        """
        assert "expired_on" in data.columns, "expired_on column is required"
        assert "start" in data.columns, "start column is required"
        assert "expiration" in data.columns, "expiration column is required"
        assert "pure_premium" in data.columns, "pure_premium column is required"
        assert "payout" in data.columns, "payout column is required"
        assert "actual_payout" in data.columns, "actual_payout column is required"

        if cols is not None:
            for col in cols:
                assert col in data.columns, f"{col} column is required"

    def review(
        self,
        show_first_date: bool = True,
        show_predicted_loss_to_exposure: bool = True,
        show_current_portfolio_pct: bool = True,
        average_duration: Optional[str] = None,
        var_levels: Optional[float | int | list[float] | list[int]] = None,
        **kwargs,
    ) -> "_CompiledReview":
        """
        Computes the portfolio review

        Parameters
        ----------
        show_first_date : bool
            whether to show the first date in the review
        show_predicted_loss_to_exposure : bool
            whether to show the predicted loss to exposure in the review
        show_current_portfolio_pct : bool
            whether to show the current portfolio percentage in the review
        average_duration : Optional[str]
            If "expected", the expected average duration is shown. If "actual", the actual average duration is shown. If None, the average duration is not shown.
        var_levels : Optional[float | list[float]]
            VaR levels to compute
        **kwargs
            arbitrary keyword arguments

        Returns
        -------
        _CompiledReview
            a compiled review object
        """

        columns = [
            "first_date",
            "pred_loss_to_exposure",
            "loss_to_exposure",
            "loss_ratio",
            "volume",
            "current_pct",
            "average_duration",
            "target-loss-to-exposure",
            "target-loss-ratio",
        ]

        if average_duration is not None:
            if average_duration not in ["expected", "actual"]:
                raise ValueError("average_duration should be 'expected', 'actual', or None")
            elif average_duration == "expected":
                self.data["duration"] = (self.data.expiration - self.data.start).dt.days
            elif average_duration == "actual":
                self.data["duration"] = (self.data.expired_on - self.data.start).dt.days

        if show_current_portfolio_pct is True:
            total_exposure = self.data.exposure.current_value()

        grouped_data = self.data.groupby(self.split_on)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            result = {
                "loss_to_exposure": grouped_data.apply(
                    lambda x: x.loss_to_exposure.current_value(post_mortem=True)
                ),
                "loss_ratio": grouped_data.apply(lambda x: x.loss_ratio.current_value(post_mortem=True)),
                "volume": grouped_data.apply(lambda x: (x.expiration <= today()).sum()),
            }
        if show_first_date is True:
            result["first_date"] = grouped_data.start.min().dt.date
        else:
            columns.remove("first_date")

        if show_current_portfolio_pct is True:
            result["current_pct"] = (
                grouped_data.apply(lambda x: x.exposure.current_value()) / total_exposure * 100
            )
        else:
            columns.remove("current_pct")

        if show_predicted_loss_to_exposure is True:
            result["pred_loss_to_exposure"] = grouped_data.apply(
                lambda x: x.pure_premium.sum() / x.payout.sum() * 100
            )
        else:
            columns.remove("pred_loss_to_exposure")

        if average_duration is not None:
            result["average_duration"] = grouped_data.duration.mean()
        else:
            columns.remove("average_duration")

        if "target_loss_ratio" in kwargs:
            target_lr_lte = grouped_data.apply(
                lambda x: pd.Series(compute_adjusted_target_loss_ratio(x, **kwargs["target_loss_ratio"]))
            )
            result["target-loss-to-exposure"] = target_lr_lte[0]
            result["target-loss-ratio"] = target_lr_lte[1]

        else:
            columns.remove("target-loss-to-exposure")
            columns.remove("target-loss-ratio")

        if var_levels is not None:
            if isinstance(var_levels, float | int):
                var_levels = [var_levels]
            var_df = grouped_data.apply(lambda x: pd.Series(_get_var(x, var_levels, **kwargs)) * 100)
            for i, l in enumerate(var_levels):
                result[f"VaR_{l}"] = var_df[i]
                columns.append(f"VaR_{l}")

        # Put the results in a single dataframe
        results = pd.DataFrame(result)[columns]
        results.sort_index(inplace=True)

        return _CompiledReview(results)


@dataclass
class _CompiledReview:
    """
    A class used to compile and represent the results of a portfolio review.

    ...

    Attributes
    ----------
    portfolio_review : pd.DataFrame
        a pandas DataFrame containing the results of the portfolio review

    Methods
    -------
    to_df() -> pd.DataFrame:
        Returns a copy of the portfolio review results as a pandas DataFrame.
    to_string(**kwargs) -> str:
        Returns a string representation of the portfolio review results.
    print(**kwargs) -> None:
        Prints the string representation of the portfolio review results.
    """

    portfolio_review: pd.DataFrame

    def to_df(self) -> pd.DataFrame:
        """
        Returns a copy of the portfolio review results as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            a copy of the portfolio review results
        """

        return self.portfolio_review.copy()

    def to_string(self, **kwargs) -> str:
        """
        Returns a string representation of the portfolio review results.

        Parameters
        ----------
        **kwargs
            arbitrary keyword arguments

        Returns
        -------
        str
            a string representation of the portfolio review results
        """

        return self.portfolio_review.to_string(float_format="{:,.2f}%".format, **kwargs)

    def print(self, **kwargs) -> None:
        """
        Prints the string representation of the portfolio review results.

        Parameters
        ----------
        **kwargs
            arbitrary keyword arguments
        """

        print(self.to_string(**kwargs))



from typing import Optional

import numpy as np
import pandas as pd

from .. import today
from ..portfolio.loss_to_exposure import current_value as loss_to_exposure_current_value
from . import ForecastResult

REQUIRED_COLUMNS = ["pure_premium", "actual_payout", "expiration", "expired_on"]


def forecast(
    data: pd.DataFrame,
    split_on: Optional[str | list[str]] = None,
    past_horizon: int = 90,
    only_on_active: bool = True,
    n_simulations: int = 1000,
    forecasted_loss_to_exposure: Optional[np.ndarray | pd.Series] = None,
) -> ForecastResult:
    """
    Forecast the loss ratio for the portfolio.
    """

    if forecasted_loss_to_exposure is not None:
        assert len(forecasted_loss_to_exposure) == len(
            data
        ), "The forecasted loss-to-exposure ratio must have the same length as the data"
        if isinstance(forecasted_loss_to_exposure, np.ndarray):
            forecasted_loss_to_exposure = pd.Series(forecasted_loss_to_exposure)
        loss_to_exposure_column = forecasted_loss_to_exposure
    else:
        loss_to_exposure_column = _forecast_loss_to_exposure_standard(data, past_horizon, split_on)

    if only_on_active is True:
        mask = ~data.actual_payout.isna()
    else:
        mask = np.ones(len(data), dtype=bool)

    loss_to_exposure_column = loss_to_exposure_column[mask].values
    payouts = data.payout[mask].values

    # Sample random numbers
    simulations = np.random.random(size=(len(loss_to_exposure_column), n_simulations))
    # Generate outcomes
    outcomes = (simulations < loss_to_exposure_column[:, np.newaxis]).astype(int)
    # Compute payouts
    outcomes = outcomes * payouts[:, np.newaxis]

    # Compute total payouts
    total_payouts = outcomes.sum(axis=0)

    # Compute total loss_ratios
    total_loss_ratios = total_payouts / data.loc[mask].pure_premium.sum()

    return ForecastResult(total_loss_ratios)


def _forecast_loss_to_exposure_standard(
    data: pd.DataFrame, past_horizon: int = 90, split_on: Optional[str | list[str]] = None
):
    """
    Forecast the loss-to-exposure ratio for the portfolio. The forecast is very simple: it assumes that the
    loss-to-exposure for each policy will be equal to the average loss-to-exposure observed in the last
    past_horizon days for the same group of policies. Policies are grouped by the columns in split_on.
    """
    if split_on is None:
        split_on = ["rm_name"]

    # First, compute the observed loss-to-exposure for the different groups
    if isinstance(split_on, str):
        split_on = [split_on]
    # Create the mask to select recent policies
    mask = data.start >= today() - pd.Timedelta(days=past_horizon)
    # Make post-mortem
    mask &= data.expiration <= today()
    loss_to_exposures = (
        data.loc[mask].groupby(split_on).apply(lambda x: loss_to_exposure_current_value(x) / 100.0).to_dict()
    )
    # Then, simulate forward
    # Create a tuple from the split_on columns to match the keys in loss_to_exposures
    if len(split_on) == 1:
        group_keys = data[split_on[0]]
    else:
        group_keys = data[split_on].apply(tuple, axis=1)
    # Map the loss_to_exposure values back to the DataFrame
    loss_to_exposure_column = group_keys.map(loss_to_exposures)
    return loss_to_exposure_column



"""
Creating pandas accessors to use the analytics functions directly from pandas
"""

import importlib
import pkgutil
import warnings
from pathlib import Path

from pandas.api.extensions import register_dataframe_accessor

# Use pathlib to get the path of the current file and then find the "metrics" subdirectory
current_dir = Path(__file__).parent
metrics_dir = current_dir / "portfolio"

function_names = ["current_value", "time_series", "at_t"]


def create_accessor(module, accessor_name):
    class CustomAccessor:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
            # Ensure the DataFrame has the necessary columns
            if not all(col in self._obj.columns for col in module.REQUIRED_COLUMNS):
                raise ValueError(f"DataFrame must contain columns: {module.REQUIRED_COLUMNS}")

    def _create_method(func_name):
        def method(self, *args, **kwargs):
            subset = self._obj[list(module.REQUIRED_COLUMNS)]
            func = getattr(module, func_name)
            return func(subset, *args, **kwargs)

        return method

    # This loop ensures each function gets its own scope with the correct func_name
    for name in function_names:
        method = _create_method(name)
        method.__name__ = name
        setattr(CustomAccessor, name, method)

    register_dataframe_accessor(accessor_name)(CustomAccessor)


def create_ensuro_accessors():
    # Catch warning "UserWaning: registration of accessor ..."
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", category=UserWarning, message="registration of accessor")
        # Iterate through the metrics and create new modules
        for _, module_name, _ in pkgutil.iter_modules([str(metrics_dir)]):
            module = importlib.import_module(f".portfolio.{module_name}", __package__)
            create_accessor(module, module_name)



import re

import pandas as pd


def date_to_period(dates, freq):
    """
    Maps datetime objects in a series to period start times based on the given frequency.
    Handles both standard and custom frequencies.

    Parameters:
    series (pd.Series): Series of datetime objects.
    freq (str): Frequency string, e.g., '10D' for ten-day periods or 'M' for months.

    Returns:
    pd.Series: Series of period start times.
    """
    # Regular expression pattern for standard frequency (e.g., 'M', 'Q', 'A', 'D', etc.)
    # Allows optional leading '1' (e.g., '1M', '1W', '1D', etc.)
    standard_freq_pattern = r"^(1)?[BQWYMDHSTLUN]+$"

    if re.match(standard_freq_pattern, freq):
        # Use pandas to_period for standard frequencies
        return dates.dt.to_period(freq).dt.start_time
    else:
        # Custom frequency handling
        # Extract the frequency number and unit (e.g., '10D' -> (10, 'D'))
        freq_num, freq_unit = int(freq[:-1]), freq[-1]

        # Create
        delta_in_days = pd.Timedelta(freq_num, unit=freq_unit).days

        # Calculate the number of units since a reference date (e.g., '1970-01-01')
        units_since_ref = (dates - pd.Timestamp("1970-01-01")).dt.days // delta_in_days

        # Calculate the start date of each custom period
        period_start_dates = pd.to_datetime(units_since_ref * freq_num, unit=freq_unit, origin="1970-01-01")

        return period_start_dates


def _max_date_range(data: pd.DataFrame, freq: str):
    """
    Computes the maximum date range (from the first start date to the last expiration date) for the given data and frequency.
    """
    start_dates = data.start.dt.normalize().unique()
    end_dates = data.expiration.dt.normalize().unique()

    dates = sorted(list(set(start_dates).union(set(end_dates))))
    dates = pd.Series(dates)

    return date_to_period(dates, freq).unique()



# -*- coding: utf-8 -*-
from os.path import abspath, dirname, join

from setuptools import find_packages, setup

this_dir = abspath(dirname(__file__))

with open(join(this_dir, "README.md"), encoding="utf-8") as file:
    long_description = file.read()

with open(join(this_dir, "requirements.txt")) as f:
    requirements = f.read().split("\n")

setup(
    name="ensuro-analytics",
    version="0.1.7",
    description="Ensuro analytics library",
    url="https://github.com/ensuro/ensuro_analytics",
    long_description_content_type="text/markdown",
    long_description=long_description,
    author="Ensuro",
    author_email="luca@ensuro.co",
    license="Apache 2.0",
    install_requires=requirements,
    packages=find_packages(exclude=["docs"]),
    include_package_data=True,
)



"""
Functions to compute the number of active policies
"""

from typing import Optional

import pandas as pd

from .. import DAY, active_at_t, today
from . import _max_date_range, date_to_period

REQUIRED_COLUMNS = ["start", "expired_on", "expiration"]


def current_value(data: pd.DataFrame, **kwargs) -> int:
    """
    Computes the number of active policies at the current date
    """
    mask = active_at_t(data, today())
    return mask.astype(int).sum()


def at_t(data: pd.DataFrame, date: pd.Timestamp, **kwargs) -> int:
    """
    Computes the number of active policies at a given date
    """
    mask = active_at_t(data, date)
    return mask.sum()


def time_series(
    data: pd.DataFrame,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: str = "start",
    **kwargs,
) -> pd.Series:
    """
    Computes the time series of active policies for the data.
    :param data: policies dataframe
    :param freq: time frequency of the time series
    :param end_date: maximum date. If None, the maximum date in the period column is used
    :param start_date: minimum date. If None, the minimum date in the period column is used
    :param period_column: can be "start", "expired_on", "expiration", or "max".
    :return:
    """
    assert "expired_on" in data.columns, "expired_on column is required"
    assert "start" in data.columns, "start column is required"

    # Get unique dates at a regular interval
    if period_column == "max":
        dates = _max_date_range(data=data, freq=freq)
    else:
        dates = sorted(date_to_period(dates=data[period_column], freq=freq).unique())

    active_policies = []
    get_data = lambda: data[["start", "expired_on"]].copy()
    # Compute active policies at each date
    for date in dates:
        active_policies.append(at_t(get_data(), date))

    active_policies = pd.Series(active_policies, index=dates, name="active_policies")

    # Filter by start and end date
    if start_date is not None:
        active_policies = active_policies.loc[active_policies.index >= start_date]
    if end_date is not None:
        active_policies = active_policies.loc[active_policies.index <= end_date]

    return active_policies



"""
Forecasting submodule. The function predict the value of a metric in the future based on historical data.
"""
from dataclasses import dataclass
from typing import Optional

import numpy as np


@dataclass
class ForecastResult:
    forecasts: np.ndarray

    @property
    def results(self):
        return self.forecasts

    @property
    def average(self):
        return np.nanmean(self.forecasts)

    @property
    def avg(self):
        return self.average

    def distribution(self, bins: Optional[int | list[float]] = 50):
        return np.histogram(self.forecasts, bins=bins)

    def median(self):
        return np.nanmedian(self.forecasts)

    def percentile(self, q: float | list[float]):
        return np.nanpercentile(self.forecasts, q)

    def p(self, q: float | list[float]):
        return self.percentile(q)

    def quantile(self, q: float | list[float]):
        return self.percentile(q)

    def q(self, q: float | list[float]):
        return self.quantile(q)



import gc
import warnings
from typing import Optional

import numpy as np
import pandas as pd

from ensuro_analytics.analytics import DAY, today

from .active_policies import at_t as active_at_
from .active_policies import current_value as active

REQUIRED_COLUMNS = ["payout", "actual_payout", "start", "expiration", "expired_on"]


def _simulation_var_level(
    _data: pd.DataFrame,
    time_window: int,
    time_col: str,
    n_policies: int,
    n_simulation: int,
    level: float | int | list[float] | list[int],
) -> float | list[float]:
    """
    Runs the simulations to compute the VaR at the given level(s).
    """

    dates = pd.date_range(start=_data[time_col].min(), end=_data[time_col].max(), freq="D")

    if time_window >= len(dates):
        dates = dates[-1:]
    else:
        dates = dates[time_window:]

    min_indices = {}
    max_indices = {}

    times = []
    for date in dates:
        tmp = _data[(_data[time_col] >= date - pd.Timedelta(days=time_window)) & (_data[time_col] < date)]
        if tmp.shape[0] > n_policies:
            times.append(date)
            min_indices[date] = tmp.index.min()
            max_indices[date] = tmp.index.max()

    times = np.array(times)

    assert len(times) > 0, "Time window is too narrow"

    # payouts and triggered policies
    payouts = _data.payout.values
    triggered = (_data.actual_payout > 0).values

    # draw which time windows bucket in each simulation
    random_times = times[np.random.choice(len(times), n_simulation, replace=True)]

    # draw which payouts
    random_policies_indices = np.zeros((n_policies, n_simulation), dtype=int)
    for j in range(n_simulation):
        # draw only policies inside the time window bucket
        random_policies_indices[:, j] = np.random.randint(
            min_indices[random_times[j]], max_indices[random_times[j]], n_policies
        )
    random_policies = payouts[random_policies_indices]

    # draw which triggered
    random_triggered_indices = np.zeros((n_policies, n_simulation), dtype=int)
    for j in range(n_simulation):
        # draw only triggers inside the time window bucket
        random_triggered_indices[:, j] = np.random.randint(
            min_indices[random_times[j]], max_indices[random_times[j]], n_policies
        )
    random_triggered = triggered[random_triggered_indices]

    # calculate the loss to exposure
    losses = np.sum(random_policies * random_triggered, axis=0)
    exp = np.sum(random_policies, axis=0)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        loss_to_exposure = losses / exp

    if isinstance(level, list):
        return [np.percentile(loss_to_exposure, l) for l in level]
    else:
        return np.percentile(loss_to_exposure, level)


def current_value(
    data: pd.DataFrame,
    level: Optional[float | int | list[float] | list[int]] = 90.0,
    n_simulation: Optional[int] = 1000,
    time_window: Optional[int] = 30,
    time_col: Optional[str] = "start",
    **kwargs,
) -> float | list[float]:
    """
    Computes the current value at risk (VaR) for the given data at the given level(s).
    Given time_window days and a time_col column, the function draws n_simulation samples of portfolio of sizes all
    equal to the currently active one, with all samples being made of policies that are within time_window days
    from each other. The function then computes the loss to exposure for each sample and returns the level-th
    percentile of the loss to exposure distribution.
    :param data: DataFrame with the data.
    :param level: VaR level(s) to compute.
    :param n_simulation: Number of simulations to run.
    :param time_window: Time window in days.
    :param time_col: Column with the time data.
    :return: VaR at the given level(s).
    """

    # number of policies to use for simulations
    n_policies = active(data)

    assert n_policies > 0, "No currently active policies"

    _data = data.copy()

    # post-mortem
    _data = _data.loc[data.expiration <= today()]

    assert (
        _data.shape[0] > n_policies
    ), "Can not compute VaR with the number of currently active portfolio: not enough historical data"

    # sort
    _data.sort_values(time_col, inplace=True)
    _data[time_col] = pd.to_datetime(_data[time_col].dt.date)
    _data.reset_index(drop=True, inplace=True)

    return _simulation_var_level(_data, time_window, time_col, n_policies, n_simulation, level)


def at_t(
    data: pd.DataFrame,
    date: pd.Timestamp,
    level: Optional[float | int | list[float] | list[int]] = 90.0,
    n_simulation: Optional[int] = 1000,
    time_window: Optional[int] = 30,
    time_col: Optional[str] = "start",
    **kwargs,
) -> float | list[float]:
    """
    Computes the value at risk (VaR) for the given data at the given level(s) at the given date.
    Given time_window days and a time_col column, the function draws n_simulation samples of policies of sizes all
    equal to the ones active at the given date, with all samples being made of portfolio that are within time_window
    days from each other. The function then computes the loss to exposure for each sample and returns the level-th
    percentile of the loss to exposure distribution.
    :param data: DataFrame with the data.
    :param date: Date at which to compute the VaR.
    :param level: VaR level(s) to compute.
    :param n_simulation: Number of simulations to run.
    :param time_window: Time window in days.
    :param time_col: Column with the time data.
    :return: VaR at the given level(s).
    """

    # number of policies to use for simulations
    n_policies = active_at_(data, date)

    assert n_policies > 0, "No active policies at the given date"

    _data = data.copy()

    # post-mortem
    _data = _data.loc[data.expiration <= date]

    assert (
        _data.shape[0] > n_policies
    ), "Can not compute VaR with the number of active portfolio at the given date: not enough historical data"

    # sort
    _data.sort_values(time_col, inplace=True)
    _data[time_col] = pd.to_datetime(_data[time_col].dt.date)
    _data.reset_index(drop=True, inplace=True)

    return _simulation_var_level(_data, time_window, time_col, n_policies, n_simulation, level)


def time_series(
    data: pd.DataFrame,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: str = "expiration",
    level: Optional[float | int | list[float] | list[int]] = 90.0,
    n_simulation: Optional[int] = 1000,
    time_window: Optional[int] = 15,
    time_col: Optional[str] = "start",
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """
    Computes the time series of the value at risk (VaR) for the given data at the given level(s).
    :param data: DataFrame with the data.
    :param freq: Frequency of the time series.
    :param end_date: End date of the time series.
    :param start_date: Start date of the time series.
    :param period_column: Column to consider as the period of the time series.
    :param level: VaR level(s) to compute.
    :param n_simulation: Number of simulations to run.
    :param time_window: Time window in days for the simulations.
    :param time_col: Column from which to draw policies in the simulations.
    :return: VaR time series at the given level(s).
    """

    _data = data.copy()

    # post-mortem
    _data = _data.loc[data.expiration <= today()]

    # sort
    _data.sort_values(time_col, inplace=True)
    _data[time_col] = pd.to_datetime(_data[time_col].dt.date)
    _data.reset_index(drop=True, inplace=True)

    # date index
    dates = pd.date_range(_data[period_column].min(), _data[period_column].max(), freq=freq)

    if isinstance(level, float | int):
        var = np.zeros(len(dates))
        for i, date in enumerate(dates):
            try:
                var[i,] = at_t(_data, date, level, n_simulation, time_window, time_col)
            except:
                var[i,] = np.nan

        var = pd.Series(var, index=dates, name="VaR")

    elif isinstance(level, list):
        var = np.zeros((len(dates), len(level)))
        for i, date in enumerate(dates):
            try:
                var[i,] = at_t(_data, date, level, n_simulation, time_window, time_col)
            except:
                var[i,] = np.nan

        var = pd.DataFrame(var, index=dates, columns=[f"VaR_{l}" for l in level])

    # Filter by start and end date
    if start_date is not None:
        var = var.loc[var.index >= start_date]
    if end_date is not None:
        var = var.loc[var.index <= end_date]

    return var



import numpy as np
import pandas as pd


def prepare_dataframe(policies, etoken_to_risk_modules_map, etoken):
    """
    Prepare the filtered dataframe with the necessary columns and calculations.
    """
    df_filter = policies.loc[policies.risk_module.isin(etoken_to_risk_modules_map.get(etoken, []))].copy()
    df_filter.expired_on = df_filter.expired_on.fillna(df_filter.expiration)
    df_filter["expired_on"] = np.minimum(df_filter["expired_on"], df_filter["expiration"])

    if df_filter.empty:
        return df_filter

    df_filter["len_first_day"] = (
        df_filter.start.dt.normalize() + pd.Timedelta(days=1) - df_filter["start"]
    ).dt.total_seconds()
    df_filter["len_last_day"] = (
        df_filter["expired_on"] - df_filter.expired_on.dt.normalize()
    ).dt.total_seconds()
    df_filter["len_last_day"] += (df_filter["expiration"] - df_filter["expired_on"]).dt.total_seconds()
    df_filter["len_other_day"] = 24 * 60 * 60
    df_filter["inv_duration_secs"] = 1 / (df_filter["expiration"] - df_filter["start"]).dt.total_seconds()

    # Handle cases where start date and expiration date are the same
    same_start_expiration = df_filter.start.dt.date == df_filter.expired_on.dt.date
    df_filter.loc[same_start_expiration, "len_first_day"] = 1.0
    df_filter.loc[same_start_expiration, "len_other_day"] = 0.0
    df_filter.loc[same_start_expiration, "len_last_day"] = 0.0
    df_filter.loc[same_start_expiration, "inv_duration_secs"] = 1.0

    for col in ["len_first_day", "len_last_day", "len_other_day"]:
        if col == "len_first_day":
            coeffs = df_filter["inv_duration_secs"].replace(np.inf, 1.0)
        else:
            coeffs = df_filter["inv_duration_secs"].replace(np.inf, 0.0)
        df_filter[col] = df_filter[col] * coeffs

    return df_filter


def compute_insurance_returns(df, tranche_col, etoken, dates):
    """
    Calculate the insurance returns for a given tranche and etoken.
    """
    results = []
    for date in dates:
        daily_return = 0
        mask_first_day = df.start.dt.date == date.date()
        mask_last_day = df.expired_on.dt.date == date.date()
        mask_other_days = (df.start.dt.date < date.date()) & (df.expired_on.dt.date > date.date())

        daily_return += (df.loc[mask_first_day, tranche_col] * df.loc[mask_first_day, "len_first_day"]).sum()
        daily_return += (df.loc[mask_last_day, tranche_col] * df.loc[mask_last_day, "len_last_day"]).sum()
        daily_return += (
            df.loc[mask_other_days, tranche_col] * df.loc[mask_other_days, "len_other_day"]
        ).sum()

        results.append(((etoken, date), daily_return))

    return results


def build_insurance_returns(data, policies, sr_etks_to_rm, jr_etks_to_rm):
    """
    Build the insurance returns for all etokens and dates.
    """
    dates = sorted(data.index.get_level_values(1).unique())
    etokens = sorted(data.index.get_level_values(0).unique())
    dates = pd.date_range(start=dates[0].normalize(), end=dates[-1].normalize(), freq="D")

    # Build the insurance returns with etokens and dates as a multi-index
    index = pd.MultiIndex.from_product([etokens, dates], names=["etoken", "date"])
    insurance_returns = pd.DataFrame(index=index, columns=["dividend_insurance"])
    insurance_returns["dividend_insurance"] = 0.0

    for etoken in etokens:
        # Returns from the senior tranche
        df_sr = prepare_dataframe(policies, sr_etks_to_rm, etoken)
        if not df_sr.empty:
            sr_returns = compute_insurance_returns(df_sr, "sr_coc", etoken, dates)
            for idx, value in sr_returns:
                insurance_returns.at[idx, "dividend_insurance"] += value

        # Returns from the junior tranche
        df_jr = prepare_dataframe(policies, jr_etks_to_rm, etoken)
        if not df_jr.empty:
            jr_returns = compute_insurance_returns(df_jr, "jr_coc", etoken, dates)
            for idx, value in jr_returns:
                insurance_returns.at[idx, "dividend_insurance"] += value

    return insurance_returns


def get_etokens_to_risk_modules_map(offchain_api):
    riskmodules = dict()
    sr_etks_to_rm = dict()
    jr_etks_to_rm = dict()
    for riskmodule in offchain_api.get("riskmodules").json():
        riskmodules[riskmodule["address"]] = dict()
        riskmodules[riskmodule["address"]]["jr_etk"] = riskmodule["jr_etk"].split("/")[-2]
        riskmodules[riskmodule["address"]]["sr_etk"] = riskmodule["sr_etk"].split("/")[-2]

        if riskmodule["sr_etk"].split("/")[-2] not in sr_etks_to_rm.keys():
            sr_etks_to_rm[riskmodule["sr_etk"].split("/")[-2]] = [riskmodule["address"]]
        else:
            sr_etks_to_rm[riskmodule["sr_etk"].split("/")[-2]].append(riskmodule["address"])

        if riskmodule["jr_etk"].split("/")[-2] not in jr_etks_to_rm.keys():
            jr_etks_to_rm[riskmodule["jr_etk"].split("/")[-2]] = [riskmodule["address"]]
        else:
            jr_etks_to_rm[riskmodule["jr_etk"].split("/")[-2]].append(riskmodule["address"])

    return riskmodules, sr_etks_to_rm, jr_etks_to_rm


def blocks_shots_to_token_metrics(
    etoken_blocks: pd.DataFrame, lps_df: pd.DataFrame, etokens_api_query: list[dict]
) -> pd.DataFrame:
    """
    Function to transform the eToken blocks shots and the LPs data into a DataFrame with daily resolution and all the necessary columns.
    :param etoken_blocks: DataFrame with the eToken blocks shots.
    :param lps_df: DataFrame with the LPs data.
    """

    # Transform the eToken blocks shots columns with correct types
    etoken_blocks["total_supply"] = etoken_blocks.total_supply.astype(float)
    etoken_blocks["scr"] = etoken_blocks.scr.astype(float)
    etoken_blocks["date"] = pd.to_datetime(etoken_blocks["date"], utc=True).dt.normalize()

    # Transform the LPs data columns with correct types
    lps_df.event_tx_timestamp = pd.to_datetime(lps_df.event_tx_timestamp)
    lps_df["date"] = lps_df.event_tx_timestamp.dt.normalize()

    # Compute the deposits and withdrawals
    lps_df["deposit"] = (lps_df.event_type == "deposit").astype(int) * lps_df.amount.astype(float)
    lps_df["withdraw"] = (lps_df.event_type == "withdraw").astype(int) * lps_df.amount.astype(float)

    # Get the eToken address and the eToken URL
    etokens_url_address_map = {r["url"]: r["address"] for r in etokens_api_query}

    # Transform the eToken URL to the eToken address
    lps_df["e_token"] = lps_df.e_token.map(etokens_url_address_map)

    # Compute the daily token balance
    daily_token_balance = (
        etoken_blocks.sort_values(["e_token", "date"])
        .groupby(["e_token", "date"])
        .agg(
            {
                "total_supply": "last",
                "scr": "last",
            }
        )
    )

    # Set a multiindex from the eToken address and the date
    e_tokens = daily_token_balance.reset_index().e_token.unique()
    dates = daily_token_balance.reset_index().date.unique()

    # Don't consider the very last date (may have incomplete data)
    dates = pd.date_range(start=dates.min(), end=dates.max(), freq="D")[:-1]

    # Set the multiindex
    reindex = pd.MultiIndex.from_product([e_tokens, dates], names=["e_token", "date"])
    daily_token_balance = daily_token_balance.reindex(reindex)

    # Fill the missing values of total supply and SCR with the last available value
    daily_token_balance = daily_token_balance.groupby(level="e_token").transform(lambda x: x.bfill().ffill())

    # Compute daily deposits and withdrawals and fill the missing values with 0
    daily_token_balance["deposit"] = lps_df.groupby(["e_token", "date"]).deposit.sum()
    daily_token_balance["withdraw"] = lps_df.groupby(["e_token", "date"]).withdraw.sum()
    daily_token_balance[["withdraw", "deposit"]] = daily_token_balance[["withdraw", "deposit"]].fillna(0)

    # Compute dividends; The dividends are computed as the difference between the total supply of the eToken
    # at time t+1 and t, minus the deposits and plus the withdrawals.
    daily_token_balance["dividend"] = (
        daily_token_balance.groupby(level="e_token").total_supply.diff()
        - daily_token_balance.deposit
        + daily_token_balance.withdraw
    )

    return daily_token_balance



"""
Functions to compute the Alpha.
"""

from typing import Optional

import numpy as np
import pandas as pd

from .. import DAY, today
from . import TimeUnits, beta, get_market_returns, risk_free_rate

REQUIRED_COLUMNS = ["returns", "perfect_returns"]


def current_value(
    data: pd.DataFrame,
    full_ur: bool = False,
    time_resolution: str | TimeUnits = "1D",
    **kwargs,
) -> float:
    """
    Function to compute the current Alpha.
    :param data: DataFrame with the eToken returns.
    :param full_ur: Boolean to indicate if we want to compute the full UR.
    :param time_resolution: time resolution of the returns’ time series used in the computation.
    :return: The Alpha.
    """

    if isinstance(time_resolution, str):
        time_resolution = TimeUnits(time_resolution)

    # Check if there are repeated dates
    assert data.index.get_level_values("date").nunique() == len(
        data.index.get_level_values("date")
    ), "There are repeated dates"

    last_date = data.index.get_level_values("date").max()

    market_data = get_market_returns(dates=data.index.get_level_values("date"))

    rf_rate = risk_free_rate(end_date=last_date, start_date=last_date - 365 * DAY)
    b = beta.current_value(data, full_ur=full_ur, time_resolution=time_resolution)

    if full_ur:
        col = "perfect_returns"
    else:
        col = "returns"

    # Compute the number of units in one year based on the time resolution
    yearly_units = time_resolution.n_units_in_one_year

    # Compute the reference rate
    ref_r = (1 + rf_rate) ** (1 / yearly_units) - 1

    returns_with_benchmarks = (
        pd.merge(market_data[["date", "returns_sp"]], data.reset_index()[["date", col]])
        .dropna()
        .set_index("date")
    )

    returns_with_benchmarks = returns_with_benchmarks[returns_with_benchmarks.notnull()]

    if len(returns_with_benchmarks) == 0:
        return np.nan

    # Compute the mean return
    returns = returns_with_benchmarks[col].mean()

    # Compute the mean S&P 500 return
    returns_mkt = returns_with_benchmarks.returns_sp.mean()

    # Compute alpha
    a = returns - (ref_r + b * (returns_mkt - ref_r))

    return a


def at_t(
    data: pd.DataFrame,
    date: pd.Timestamp,
    full_ur: bool = False,
    time_resolution: str | TimeUnits = "1D",
    **kwargs,
) -> float:
    """
    Function to compute the Alpha at a given date.
    :param data: DataFrame with the eToken returns.
    :param date: Date at which we want to compute the temporal weighted yearly return.
    :param full_ur: Boolean to indicate if we want to compute the full UR.
    :param time_resolution:  time resolution of the returns’ time series used in the computation.
    :return: The Alpha at the given date.
    """

    if date.tz is None:
        date = date.tz_localize("UTC")

    if isinstance(time_resolution, str):
        time_resolution = TimeUnits(time_resolution)

    assert date > data.index.get_level_values("date").min() + pd.Timedelta(
        f"{time_resolution.n_days_in_unit}D"
    ), "Date is too early"

    mask = data.index.get_level_values("date") < date

    return current_value(data.loc[mask], full_ur=full_ur, time_resolution=time_resolution)


def time_series(
    data: pd.DataFrame,
    full_ur: bool = False,
    time_resolution: str | TimeUnits = "1D",
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    **kwargs,
) -> pd.Series:
    """
    Computes the Alpha.
    :param data: DataFrame with the eToken returns.
    :param full_ur: Boolean to indicate if we want to compute the full UR.
    :param time_resolution: time resolution of the returns’ time series used in the computation.
    :param freq: Frequency of the time series.
    :param end_date: Maximum date. If None, the maximum date in the period column is used.
    :param start_date: Minimum date. If None, the minimum date in the period column is used.
    :return: The Alpha time series.
    """

    # Localize the dates as UTC if they are not localized
    if start_date.tz is None:
        start_date = start_date.tz_localize("UTC")

    if end_date.tz is None:
        end_date = end_date.tz_localize("UTC")

    if isinstance(time_resolution, str):
        time_resolution = TimeUnits(time_resolution)

    freq_unit = TimeUnits("1" + freq[-1])
    freq_days = int(freq[:-1]) * freq_unit.n_days_in_unit

    start_date = max(
        start_date,
        data.index.get_level_values("date").min() + pd.Timedelta(f"{freq_days}D"),
        data.index.get_level_values("date").min() + pd.Timedelta(f"{time_resolution.n_days_in_unit}D"),
    )

    ts_dates = pd.date_range(start=start_date, end=end_date, freq=freq, tz="UTC")

    ts = pd.Series(
        index=ts_dates,
        data=[at_t(data, date, full_ur=full_ur, time_resolution=time_resolution) for date in ts_dates],
    )

    return ts



"""
Functions to compute the senior solvency capital
"""

from typing import Optional

import pandas as pd

from .. import DAY, active_at_t, today
from . import _max_date_range, date_to_period

REQUIRED_COLUMNS = ["sr_scr", "expired_on", "start", "expiration"]


def current_value(data: pd.DataFrame, **kwargs) -> float:
    """
    Computes the current value of the active Senior solvency capital locked
    """
    mask = active_at_t(data, pd.to_datetime("today"))
    return data.loc[mask].sr_scr.sum()


def at_t(data: pd.DataFrame, date: pd.Timestamp, **kwargs) -> float:
    """
    Computes the value of the active Senior solvency capital locked at a given date
    """
    mask = active_at_t(data, date)
    return data.loc[mask].sr_scr.sum()


def time_series(
    data: pd.DataFrame,
    freq: str = "1W",
    end_date: Optional[pd.Timestamp] = today(),
    start_date: Optional[pd.Timestamp] = today() - 90 * DAY,
    period_column: str = "start",
    **kwargs,
) -> pd.Series:
    """
    Computes the time series of active senior solvency capital locked for the data.
    :param data: policies dataframe.
    :param freq: time series frequency
    :param end_date: maximum date. If None, the maximum date in the period column is used
    :param start_date: minimum date. If None, the minimum date in the period column is used
    :param period_column: can be "start", "expiration", "expired_on".
    """

    # Get dates at regular intervals
    if period_column == "max":
        dates = _max_date_range(data=data, freq=freq)
    else:
        dates = sorted(date_to_period(dates=data[period_column], freq=freq).unique())

    sr_scr = []

    get_data = lambda: data[["start", "expired_on", "expiration", "sr_scr"]].copy()

    for date in dates:
        sr_scr.append(at_t(get_data(), date))

    sr_scr = pd.Series(sr_scr, index=dates, name="jr_scr")

    if start_date is not None:
        sr_scr = sr_scr.loc[sr_scr.index >= start_date]
    if end_date is not None:
        sr_scr = sr_scr.loc[sr_scr.index <= end_date]

    return sr_scr
